# coding=utf-8
"""
choose.py - Sopel Choice Plugin
Copyright 2010-2013, Dimitri "Tyrope" Molenaars, TyRope.nl
Copyright 2013, Ari Koivula, <ari@koivu.la>
Copyright 2018, Florian Strzelecki, <florian.strzelecki@gmail.com>
Copyright 2019, dgw, technobabbl.es
Licensed under the Eiffel Forum License 2.

https://sopel.chat
"""
from __future__ import absolute_import, division, print_function, unicode_literals

import random
import unicodedata

from sopel import formatting, plugin

# Remove when dropping py2 support
try:
    str = unicode
except NameError:
    pass


def _format_safe(text):
    """Remove excess whitespace and terminate IRC formatting.

    :param str text: text to clean of whitespace
    :rtype: str
    :raises TypeError: if the passed ``text`` is not a string

    Our own take on ``str.strip()`` that skips stripping off IRC formatting
    and makes sure any formatting codes are closed if necessary.
    """
    if not isinstance(text, str):
        raise TypeError("A string is required.")
    elif not text:
        # unnecessary optimization
        return ''

    start = end = 0

    # strip left
    pos = 0
    while pos < len(text):
        is_whitespace = unicodedata.category(text[pos]) == 'Zs'
        is_non_printing = (
            text[pos] in formatting.CONTROL_NON_PRINTING and
            text[pos] not in formatting.CONTROL_FORMATTING
        )
        if not is_whitespace and not is_non_printing:
            start = pos
            break
        pos += 1

    # strip right
    pos = len(text) - 1
    while pos >= 0:
        is_whitespace = unicodedata.category(text[pos]) == 'Zs'
        is_non_printing = (
            text[pos] in formatting.CONTROL_NON_PRINTING and
            text[pos] not in formatting.CONTROL_FORMATTING
        )
        if not is_whitespace and not is_non_printing:
            end = pos + 1
            break
        pos -= 1

    # build the final string
    safe = text[start:end]
    if any(c in safe for c in formatting.CONTROL_FORMATTING):
        # if it includes IRC formatting, append reset character just in case
        safe += formatting.CONTROL_NORMAL

    return safe


@plugin.command('choose', 'choice', 'ch')
@plugin.priority("medium")
@plugin.example(".choose a, b, c", r'Your options: a, b, c. My choice: (a|b|c)', re=True)
@plugin.example(".choose a | b | c", r'Your options: a, b, c. My choice: (a|b|c)', re=True)
@plugin.example(".choose a,b,c", r'Your options: a, b, c. My choice: (a|b|c)', re=True)
@plugin.example(".choose a|b|c", r'Your options: a, b, c. My choice: (a|b|c)', re=True)
@plugin.example(".choose a b c", r'Your options: a, b, c. My choice: (a|b|c)', re=True)
@plugin.example(".choose a, b | just a",
                r'Your options: "a, b", just a. My choice: ((a, b)|(just a))',
                re=True)
@plugin.example(".choose a", 'Your options: a. My choice: a')
@plugin.example(".choose a | b | c", user_help=True)
@plugin.example(".choose a, b, c", user_help=True)
def choose(bot, trigger):
    """Makes a difficult choice easy."""
    if not trigger.group(2):
        bot.reply("I'd choose an option, but you didn't give me any.")
        return

    choices = [trigger.group(2)]
    for delim in '|\\/, ':
        choices = trigger.group(2).split(delim)
        if len(choices) > 1:
            break
    choices = [_format_safe(choice) for choice in choices]
    pick = random.choice(choices)

    # Always use a comma in the output
    display_options = ', '.join(
        choice if ',' not in choice else '"%s"' % choice
        for choice in choices
    )
    bot.reply('Your options: %s. My choice: %s' % (display_options, pick))



# coding=utf-8
# Copyright 2019, Florian Strzelecki <florian.strzelecki@gmail.com>
#
# Licensed under the Eiffel Forum License 2.
# When working on core IRC protocol related features, consult protocol
# documentation at http://www.irchelp.org/irchelp/rfc/
from __future__ import absolute_import, division, print_function, unicode_literals

import asynchat
import asyncore
import datetime
import errno
import logging
import os
import socket
import sys
import threading

from sopel import loader, plugin
from sopel.tools import jobs
from .abstract_backends import AbstractIRCBackend
from .utils import get_cnames

try:
    import ssl
    if not hasattr(ssl, 'match_hostname'):
        # Attempt to import ssl_match_hostname from python-backports
        # TODO: Remove when dropping Python 2 support
        import backports.ssl_match_hostname
        ssl.match_hostname = backports.ssl_match_hostname.match_hostname
        ssl.CertificateError = backports.ssl_match_hostname.CertificateError
    has_ssl = True
except ImportError:
    # no SSL support
    has_ssl = False

if sys.version_info.major >= 3:
    unicode = str


LOGGER = logging.getLogger(__name__)


@plugin.thread(False)
@plugin.interval(5)
def _send_ping(backend):
    if not backend.is_connected():
        return

    events = []
    need_ping = True

    # Ensure we have a time to check first
    if backend.last_event_at:
        events.append(backend.last_event_at)

    if backend.last_ping_at:
        events.append(backend.last_ping_at)

    # At least a PING was sent, or a message was received
    if events:
        last_event = max(events)
        dt = datetime.datetime.utcnow() - last_event
        time_passed = dt.total_seconds()
        need_ping = time_passed > backend.ping_interval

    # Send PING only if needed
    if need_ping:
        try:
            backend.send_ping(backend.host)
            backend.last_ping_at = datetime.datetime.utcnow()
        except socket.error:
            LOGGER.exception('Socket error on PING')


@plugin.thread(False)
@plugin.interval(10)
def _check_timeout(backend):
    if not backend.is_connected():
        return
    dt = datetime.datetime.utcnow() - backend.last_event_at
    time_passed = dt.total_seconds()
    if time_passed > backend.server_timeout:
        LOGGER.error(
            'Server timeout detected after %ss; closing.', time_passed)
        # discard buffers: no need to read/write anything more, just quit
        LOGGER.debug('Discard current buffers.')
        backend.discard_buffers()
        # close now
        backend.handle_close()


class AsynchatBackend(AbstractIRCBackend, asynchat.async_chat):
    """IRC backend implementation using :mod:`asynchat` (:mod:`asyncore`).

    :param bot: a Sopel instance
    :type bot: :class:`sopel.bot.Sopel`
    :param int server_timeout: connection timeout in seconds
    :param int ping_interval: ping interval in seconds

    The ``server_timeout`` option defaults to ``120`` seconds if not provided.

    The ``ping_interval`` defaults to ``server_timeout * 0.45`` if not specified.
    """
    def __init__(self, bot, server_timeout=None, ping_interval=None, **kwargs):
        AbstractIRCBackend.__init__(self, bot)
        asynchat.async_chat.__init__(self)
        self.writing_lock = threading.RLock()
        self.set_terminator(b'\n')
        self.buffer = ''
        self.server_timeout = server_timeout or 120
        self.ping_interval = ping_interval or (self.server_timeout * 0.45)
        self.last_event_at = None
        self.last_ping_at = None
        self.host = None
        self.port = None
        self.source_address = None
        self.timeout_scheduler = jobs.Scheduler(self)

        # register timeout jobs
        self.register_timeout_jobs([
            _send_ping,
            _check_timeout,
        ])

    def is_connected(self):
        return self.connected

    def on_irc_error(self, pretrigger):
        if self.bot.hasquit:
            # discard buffers: no need to read/write anything more, just quit
            LOGGER.debug('Discard current buffers.')
            self.discard_buffers()
            # close now
            self.handle_close()

    def irc_send(self, data):
        """Send an IRC line as raw ``data`` to the socket connection.

        :param bytes data: raw line to send

        This uses :meth:`asyncore.dispatcher.send` method to send ``data``
        directly. This method is thread-safe.
        """
        with self.writing_lock:
            self.send(data)

    def run_forever(self):
        """Run forever."""
        LOGGER.debug('Running forever.')
        asyncore.loop()

    def register_timeout_jobs(self, handlers):
        """Register the timeout handlers for the timeout scheduler."""
        for handler in handlers:
            loader.clean_callable(handler, self.bot.settings)
            job = jobs.Job.from_callable(self.bot.settings, handler)
            self.timeout_scheduler.register(job)
            LOGGER.debug('Timeout Job registered: %s', str(job))

    def initiate_connect(self, host, port, source_address):
        """Initiate IRC connection.

        :param str host: IRC server hostname
        :param int port: IRC server port
        :param str source_address: the source address from which to initiate
                                   the connection attempt
        """
        self.host = host
        self.port = port
        self.source_address = source_address

        LOGGER.info('Connecting to %s:%s...', host, port)
        try:
            LOGGER.debug('Set socket')
            self.set_socket(socket.create_connection((host, port),
                            source_address=source_address))
            LOGGER.debug('Connection attempt')
            self.connect((host, port))
        except socket.error as e:
            LOGGER.exception('Connection error: %s', e)
            self.handle_close()

    def handle_connect(self):
        """Called when the active opener's socket actually makes a connection."""
        LOGGER.info('Connection accepted by the server...')
        self.timeout_scheduler.start()
        self.bot.on_connect()

    def handle_close(self):
        """Called when the connection must be closed."""
        LOGGER.debug('Stopping timeout watchdog')
        self.timeout_scheduler.stop()
        LOGGER.info('Closing connection')
        self.close()
        self.bot.on_close()

    def handle_error(self):
        """Called when an exception is raised and not otherwise handled.

        This method is an override of :meth:`asyncore.dispatcher.handle_error`,
        the :class:`asynchat.async_chat` being a subclass of
        :class:`asyncore.dispatcher`.
        """
        LOGGER.info('Connection error...')
        self.bot.on_error()

    def collect_incoming_data(self, data):
        """Try to make sense of incoming data as Unicode.

        :param bytes data: the incoming raw bytes

        The incoming line is discarded (and thus ignored) if guessing the text
        encoding and decoding it fails.
        """
        # We can't trust clients to pass valid Unicode.
        try:
            data = unicode(data, encoding='utf-8')
        except UnicodeDecodeError:
            # not Unicode; let's try CP-1252
            try:
                data = unicode(data, encoding='cp1252')
            except UnicodeDecodeError:
                # Okay, let's try ISO 8859-1
                try:
                    data = unicode(data, encoding='iso8859-1')
                except UnicodeDecodeError:
                    # Discard line if encoding is unknown
                    return
        if data:
            self.bot.log_raw(data, '<<')
        self.buffer += data
        self.last_event_at = datetime.datetime.utcnow()

    def found_terminator(self):
        """Handle the end of an incoming message."""
        line = self.buffer
        if line.endswith('\r'):
            line = line[:-1]
        self.buffer = ''
        self.bot.on_message(line)

    def on_scheduler_error(self, scheduler, exc):
        """Called when the Job Scheduler fails."""
        LOGGER.exception('Error with the timeout scheduler: %s', exc)
        self.handle_close()

    def on_job_error(self, scheduler, job, exc):
        """Called when a job from the Job Scheduler fails."""
        LOGGER.exception('Error with the timeout scheduler: %s', exc)
        self.handle_close()


class SSLAsynchatBackend(AsynchatBackend):
    """SSL-aware extension of :class:`AsynchatBackend`.

    :param bot: a Sopel instance
    :type bot: :class:`sopel.bot.Sopel`
    :param bool verify_ssl: whether to validate the IRC server's certificate
                            (default ``True``, for good reason)
    :param str ca_certs: filesystem path to a CA Certs file containing trusted
                         root certificates
    """
    def __init__(self, bot, verify_ssl=True, ca_certs=None, **kwargs):
        AsynchatBackend.__init__(self, bot, **kwargs)
        self.verify_ssl = verify_ssl
        self.ssl = None
        self.ca_certs = ca_certs

    def handle_connect(self):
        """Handle potential TLS connection."""
        # TODO: Refactor to use SSLContext and an appropriate PROTOCOL_* constant
        # See https://lgtm.com/rules/1507225275976/
        # These warnings are ignored for now, because we can't easily fix them
        # while maintaining compatibility with py2.7 AND 3.3+, but in Sopel 8
        # the supported range should narrow sufficiently to fix these for real.
        # Each Python version still generally selects the most secure protocol
        # version(s) it supports.
        if not self.verify_ssl:
            self.ssl = ssl.wrap_socket(self.socket,  # lgtm [py/insecure-default-protocol]
                                       do_handshake_on_connect=True,
                                       suppress_ragged_eofs=True)
        else:
            self.ssl = ssl.wrap_socket(self.socket,  # lgtm [py/insecure-default-protocol]
                                       do_handshake_on_connect=True,
                                       suppress_ragged_eofs=True,
                                       cert_reqs=ssl.CERT_REQUIRED,
                                       ca_certs=self.ca_certs)
            # connect to host specified in config first
            try:
                ssl.match_hostname(self.ssl.getpeercert(), self.host)
            except ssl.CertificateError:
                # the host in config and certificate don't match
                LOGGER.error("hostname mismatch between configuration and certificate")
                # check (via exception) if a CNAME matches as a fallback
                has_matched = False
                for hostname in get_cnames(self.host):
                    try:
                        ssl.match_hostname(self.ssl.getpeercert(), hostname)
                        LOGGER.warning(
                            "using {0} instead of {1} for TLS connection"
                            .format(hostname, self.host))
                        has_matched = True
                        break
                    except ssl.CertificateError:
                        pass

                if not has_matched:
                    # everything is broken
                    LOGGER.error("Invalid certificate, no hostname matches.")
                    # TODO: refactor access to bot's settings
                    if hasattr(self.bot.settings.core, 'pid_file_path'):
                        # TODO: refactor to quit properly (no "os._exit")
                        os.unlink(self.bot.settings.core.pid_file_path)
                        os._exit(1)
        self.set_socket(self.ssl)
        LOGGER.info('Connection accepted by the server...')
        LOGGER.debug('Starting job scheduler for connection timeout...')
        self.timeout_scheduler.start()
        self.bot.on_connect()

    def send(self, data):
        """SSL-aware override for :meth:`~asyncore.dispatcher.send`."""
        try:
            result = self.socket.send(data)
            return result
        except ssl.SSLError as why:
            if why[0] in (asyncore.EWOULDBLOCK, errno.ESRCH):
                return 0
            raise why

    def recv(self, buffer_size):
        """SSL-aware override for :meth:`~asyncore.dispatcher.recv`.

        From a (now deleted) blog post by Evan "K7FOS" Fosmark:
        https://k7fos.com/2010/09/ssl-support-in-asynchatasync_chat
        """
        try:
            data = self.socket.read(buffer_size)
            if not data:
                self.handle_close()
                return b''
            return data
        except ssl.SSLError as why:
            if why[0] in (asyncore.ECONNRESET, asyncore.ENOTCONN,
                          asyncore.ESHUTDOWN):
                self.handle_close()
                return ''
            elif why[0] == errno.ENOENT:
                # Required in order to keep it non-blocking
                return b''
            else:
                raise



# coding=utf-8
# ASCII ONLY IN THIS FILE THOUGH!!!!!!!
# Python does some stupid bullshit of respecting LC_ALL over the encoding on the
# file, so in order to undo Python's ridiculous fucking idiocy, we have to have
# our own check.

# Copyright 2008, Sean B. Palmer, inamidst.com
# Copyright 2012, Elsie Powell, http://embolalia.com
# Copyright 2012, Elad Alfassa <elad@fedoraproject.org>
#
# Licensed under the Eiffel Forum License 2.

from __future__ import absolute_import, division, print_function, unicode_literals

from collections import namedtuple
import locale
import re
import sys

import pkg_resources

__all__ = [
    'bot',
    'config',
    'db',
    'formatting',
    'irc',
    'loader',
    'logger',
    'module',  # deprecated in 7.1, removed in 9.0
    'plugin',
    'tools',
    'trigger',
    'version_info',
]

loc = locale.getlocale()
if sys.version_info.major > 2:
    if not loc[1] or ('UTF-8' not in loc[1] and 'utf8' not in loc[1]):
        print('WARNING!!! You are running with a non-UTF8 locale environment '
              'variables (e.g. LC_ALL is set to "C"), which makes Python 3 do '
              'stupid things. If you get strange errors, please set it to '
              'something like "en_US.UTF-8".', file=sys.stderr)


__version__ = pkg_resources.get_distribution('sopel').version


def _version_info(version=__version__):
    regex = re.compile(r'(\d+)\.(\d+)\.(\d+)(?:[\-\.]?(a|b|rc)(\d+))?.*')
    version_groups = regex.match(version).groups()
    major, minor, micro = (int(piece) for piece in version_groups[0:3])
    level = version_groups[3]
    serial = int(version_groups[4] or 0)
    if level == 'a':
        level = 'alpha'
    elif level == 'b':
        level = 'beta'
    elif level == 'rc':
        level = 'candidate'
    elif not level and version_groups[4] is None:
        level = 'final'
    else:
        level = 'alpha'
    version_type = namedtuple('version_info',
                              'major, minor, micro, releaselevel, serial')
    return version_type(major, minor, micro, level, serial)


version_info = _version_info()



# coding=utf-8
"""Test factories: they create objects for testing purposes.

.. versionadded:: 7.0
"""
from __future__ import absolute_import, division, print_function, unicode_literals

import re

from sopel import bot, config, plugins, trigger
from .mocks import MockIRCBackend, MockIRCServer, MockUser


class BotFactory(object):
    """Factory to create bot.

    .. seealso::

        The :func:`~sopel.tests.pytest_plugin.botfactory` fixture can be used
        to instantiate this factory.
    """
    def preloaded(self, settings, preloads=None):
        """Create a bot and preload its plugins.

        :param settings: Sopel's configuration for testing purposes
        :type settings: :class:`sopel.config.Config`
        :param list preloads: list of plugins to preload, setup, and register
        :return: a test instance of the bot
        :rtype: :class:`sopel.bot.Sopel`

        This will instantiate a :class:`~sopel.bot.Sopel` object, replace its
        backend with a :class:`~.mocks.MockIRCBackend`, and then preload
        plugins. This will automatically load the ``coretasks`` plugin, and
        every other plugin from ``preloads``::

            factory = BotFactory()
            bot = factory.preloaded(settings, ['emoticons', 'remind'])

        .. note::

            This will automatically setup plugins: be careful with plugins that
            require access to external services on setup.

            You may also need to manually call shutdown routines for the
            loaded plugins.

        """
        preloads = set(preloads or []) | {'coretasks'}
        mockbot = self(settings)

        usable_plugins = plugins.get_usable_plugins(settings)
        for name in preloads:
            plugin = usable_plugins[name][0]
            plugin.load()
            plugin.setup(mockbot)
            plugin.register(mockbot)

        return mockbot

    def __call__(self, settings):
        obj = bot.Sopel(settings, daemon=False)
        obj.backend = MockIRCBackend(obj)
        return obj


class ConfigFactory(object):
    """Factory to create settings.

    .. seealso::

        The :func:`~sopel.tests.pytest_plugin.configfactory` fixture can be
        used to instantiate this factory.
    """
    def __init__(self, tmpdir):
        self.tmpdir = tmpdir

    def __call__(self, name, data):
        tmpfile = self.tmpdir.join(name)
        tmpfile.write(data)
        return config.Config(tmpfile.strpath)


class TriggerFactory(object):
    """Factory to create trigger.

    .. seealso::

        The :func:`~sopel.tests.pytest_plugin.triggerfactory` fixture can be
        used to instantiate this factory.
    """
    def wrapper(self, mockbot, raw, pattern=None):
        trigger = self(mockbot, raw, pattern=pattern)
        return bot.SopelWrapper(mockbot, trigger)

    def __call__(self, mockbot, raw, pattern=None):
        url_schemes = mockbot.settings.core.auto_url_schemes
        return trigger.Trigger(
            mockbot.settings,
            trigger.PreTrigger(mockbot.nick, raw, url_schemes=url_schemes),
            re.match(pattern or r'.*', raw))


class IRCFactory(object):
    """Factory to create mock IRC server.

    .. seealso::

        The :func:`~sopel.tests.pytest_plugin.ircfactory` fixture can be used
        to create this factory.
    """
    def __call__(self, mockbot, join_threads=True):
        return MockIRCServer(mockbot, join_threads)


class UserFactory(object):
    """Factory to create mock user.

    .. seealso::

        The :func:`~sopel.tests.pytest_plugin.userfactory` fixture can be used
        to create this factory.
    """
    def __call__(self, nick=None, user=None, host=None):
        return MockUser(nick, user, host)



# coding=utf-8
"""
meetbot.py - Sopel Meeting Logger Plugin
This plugin is an attempt to implement some of the functionality of Debian's meetbot
Copyright © 2012, Elad Alfassa, <elad@fedoraproject.org>
Licensed under the Eiffel Forum License 2.

https://sopel.chat
"""
from __future__ import absolute_import, division, print_function, unicode_literals

import codecs
import collections
import os
import re
from string import punctuation, whitespace
import time

from sopel import formatting, plugin, tools
from sopel.config import types
from sopel.modules.url import find_title


UNTITLED_MEETING = "Untitled meeting"


class MeetbotSection(types.StaticSection):
    """Configuration file section definition"""

    meeting_log_path = types.FilenameAttribute(
        "meeting_log_path",
        relative=False,
        directory=True,
        default="~/www/meetings")
    """Path to meeting logs storage directory.

    This should be an absolute path, accessible on a webserver.
    """

    meeting_log_baseurl = types.ValidatedAttribute(
        "meeting_log_baseurl",
        default="http://localhost/~sopel/meetings")
    """Base URL for the meeting logs directory."""


def configure(config):
    """
    | name | example | purpose |
    | ---- | ------- | ------- |
    | meeting\\_log\\_path | /home/sopel/www/meetings | Path to meeting logs storage directory (should be an absolute path, accessible on a webserver) |
    | meeting\\_log\\_baseurl | http://example.com/~sopel/meetings | Base URL for the meeting logs directory |
    """
    config.define_section("meetbot", MeetbotSection)
    config.meetbot.configure_setting(
        "meeting_log_path", "Enter the directory to store logs in."
    )
    config.meetbot.configure_setting(
        "meeting_log_baseurl", "Enter the base URL for the meeting logs."
    )


def setup(bot):
    bot.config.define_section("meetbot", MeetbotSection)


meetings_dict = collections.defaultdict(dict)  # Saves metadata about currently running meetings
"""
meetings_dict is a 2D dict.

Each meeting should have:
channel
time of start
head (can stop the meeting, plus all abilities of chairs)
chairs (can add infolines to the logs)
title
current subject
comments (what people who aren't voiced want to add)

Using channel as the meeting ID as there can't be more than one meeting in a
channel at the same time.
"""

# To be defined on meeting start as part of sanity checks, used by logging
# functions so we don't have to pass them bot
meeting_log_path = ""
meeting_log_baseurl = ""

# A dict of channels to the actions that have been created in them. This way
# we can have .listactions spit them back out later on.
meeting_actions = {}


# Get the logfile name for the meeting in the requested channel
# Used by all logging functions and web path
def figure_logfile_name(channel):
    if meetings_dict[channel]["title"] == UNTITLED_MEETING:
        name = "untitled"
    else:
        name = meetings_dict[channel]["title"]
    # Real simple sluggifying.
    # May not handle unicode or unprintables well. Close enough.
    for character in punctuation + whitespace:
        name = name.replace(character, "-")
    name = name.strip("-")
    timestring = time.strftime(
        "%Y-%m-%d-%H:%M", time.gmtime(meetings_dict[channel]["start"])
    )
    filename = timestring + "_" + name
    return filename


# Start HTML log
def log_html_start(channel):
    logfile_filename = os.path.join(
        meeting_log_path + channel, figure_logfile_name(channel) + ".html"
    )
    logfile = codecs.open(logfile_filename, "a", encoding="utf-8")
    timestring = time.strftime(
        "%Y-%m-%d %H:%M", time.gmtime(meetings_dict[channel]["start"])
    )
    title = "%s at %s, %s" % (meetings_dict[channel]["title"], channel, timestring)
    logfile.write(
        (
            "<!doctype html><html><head><meta charset='utf-8'>\n"
            "<title>{title}</title>\n</head><body>\n<h1>{title}</h1>\n"
        ).format(title=title)
    )
    logfile.write(
        "<h4>Meeting started by %s</h4><ul>\n" % meetings_dict[channel]["head"]
    )
    logfile.close()


# Write a list item in the HTML log
def log_html_listitem(item, channel):
    logfile_filename = os.path.join(
        meeting_log_path + channel, figure_logfile_name(channel) + ".html"
    )
    logfile = codecs.open(logfile_filename, "a", encoding="utf-8")
    logfile.write("<li>" + item + "</li>\n")
    logfile.close()


# End the HTML log
def log_html_end(channel):
    logfile_filename = os.path.join(
        meeting_log_path + channel, figure_logfile_name(channel) + ".html"
    )
    logfile = codecs.open(logfile_filename, "a", encoding="utf-8")
    current_time = time.strftime("%H:%M:%S", time.gmtime())
    logfile.write("</ul>\n<h4>Meeting ended at %s UTC</h4>\n" % current_time)
    plainlog_url = meeting_log_baseurl + tools.web.quote(
        channel + "/" + figure_logfile_name(channel) + ".log"
    )
    logfile.write('<a href="%s">Full log</a>' % plainlog_url)
    logfile.write("\n</body>\n</html>\n")
    logfile.close()


# Write a string to the plain text log
def log_plain(item, channel):
    logfile_filename = os.path.join(
        meeting_log_path + channel, figure_logfile_name(channel) + ".log"
    )
    logfile = codecs.open(logfile_filename, "a", encoding="utf-8")
    current_time = time.strftime("%H:%M:%S", time.gmtime())
    logfile.write("[" + current_time + "] " + item + "\r\n")
    logfile.close()


# Check if a meeting is currently running
def is_meeting_running(channel):
    try:
        return meetings_dict[channel]["running"]
    except KeyError:
        return False


# Check if nick is a chair or head of the meeting
def is_chair(nick, channel):
    try:
        return (
            nick.lower() == meetings_dict[channel]["head"] or
            nick.lower() in meetings_dict[channel]["chairs"]
        )
    except KeyError:
        return False


# Start meeting (also performs all required sanity checks)
@plugin.command("startmeeting")
@plugin.example(".startmeeting", user_help=True)
@plugin.example(".startmeeting Meeting Title", user_help=True)
@plugin.require_chanmsg("Meetings can only be started in channels")
def startmeeting(bot, trigger):
    """
    Start a meeting.\
    See [meetbot plugin usage]({% link _usage/meetbot-plugin.md %})
    """
    if is_meeting_running(trigger.sender):
        bot.say("There is already an active meeting here!")
        return
    # Start the meeting
    meetings_dict[trigger.sender]["start"] = time.time()
    if not trigger.group(2):
        meetings_dict[trigger.sender]["title"] = UNTITLED_MEETING
    else:
        meetings_dict[trigger.sender]["title"] = trigger.group(2)
    meetings_dict[trigger.sender]["head"] = trigger.nick.lower()
    meetings_dict[trigger.sender]["running"] = True
    meetings_dict[trigger.sender]["comments"] = []

    # Set up paths and URLs
    global meeting_log_path
    meeting_log_path = bot.config.meetbot.meeting_log_path
    if not meeting_log_path.endswith(os.sep):
        meeting_log_path += os.sep

    global meeting_log_baseurl
    meeting_log_baseurl = bot.config.meetbot.meeting_log_baseurl
    if not meeting_log_baseurl.endswith("/"):
        meeting_log_baseurl = meeting_log_baseurl + "/"

    channel_log_path = meeting_log_path + trigger.sender
    if not os.path.isdir(channel_log_path):
        try:
            os.makedirs(channel_log_path)
        except Exception:  # TODO: Be specific
            bot.say(
                "Meeting not started: Couldn't create log directory for this channel"
            )
            meetings_dict[trigger.sender] = collections.defaultdict(dict)
            raise
    # Okay, meeting started!
    log_plain("Meeting started by " + trigger.nick.lower(), trigger.sender)
    log_html_start(trigger.sender)
    meeting_actions[trigger.sender] = []
    bot.say(
        (
            formatting.bold("Meeting started!") + " use {0}action, {0}agreed, "
            "{0}info, {0}link, {0}chairs, {0}subject, and {0}comments to "
            "control the meeting. To end the meeting, type {0}endmeeting"
        ).format(bot.config.core.help_prefix)
    )
    bot.say(
        (
            "Users without speaking permission can participate by sending me "
            "a PM with `{0}comment {1}` followed by their comment."
        ).format(bot.config.core.help_prefix, trigger.sender)
    )


# Change the current subject (will appear as <h3> in the HTML log)
@plugin.command("subject")
@plugin.example(".subject roll call")
def meetingsubject(bot, trigger):
    """
    Change the meeting subject.\
    See [meetbot plugin usage]({% link _usage/meetbot-plugin.md %})
    """
    if not is_meeting_running(trigger.sender):
        bot.say("There is no active meeting")
        return
    if not trigger.group(2):
        bot.say("What is the subject?")
        return
    if not is_chair(trigger.nick, trigger.sender):
        bot.say("Only meeting head or chairs can do that")
        return
    meetings_dict[trigger.sender]["current_subject"] = trigger.group(2)
    logfile_filename = os.path.join(
        meeting_log_path + trigger.sender, figure_logfile_name(trigger.sender) + ".html"
    )
    logfile = codecs.open(logfile_filename, "a", encoding="utf-8")
    logfile.write("</ul><h3>" + trigger.group(2) + "</h3><ul>")
    logfile.close()
    log_plain(
        "Current subject: {} (set by {})".format(trigger.group(2), trigger.nick),
        trigger.sender,
    )
    bot.say(formatting.bold("Current subject:") + " " + trigger.group(2))


# End the meeting
@plugin.command("endmeeting")
@plugin.example(".endmeeting")
def endmeeting(bot, trigger):
    """
    End a meeting.\
    See [meetbot plugin usage]({% link _usage/meetbot-plugin.md %})
    """
    if not is_meeting_running(trigger.sender):
        bot.say("There is no active meeting")
        return
    if not is_chair(trigger.nick, trigger.sender):
        bot.say("Only meeting head or chairs can do that")
        return
    meeting_length = time.time() - meetings_dict[trigger.sender]["start"]
    bot.say(
        formatting.bold("Meeting ended!") +
        " Total meeting length %d minutes" % (meeting_length // 60)
    )
    log_html_end(trigger.sender)
    htmllog_url = meeting_log_baseurl + tools.web.quote(
        trigger.sender + "/" + figure_logfile_name(trigger.sender) + ".html"
    )
    log_plain(
        "Meeting ended by %s. Total meeting length: %d minutes"
        % (trigger.nick, meeting_length // 60),
        trigger.sender,
    )
    bot.say("Meeting minutes: " + htmllog_url)
    meetings_dict[trigger.sender] = collections.defaultdict(dict)
    del meeting_actions[trigger.sender]


# Set meeting chairs (people who can control the meeting)
@plugin.command("chairs")
@plugin.example(".chairs Tyrope Jason elad")
def chairs(bot, trigger):
    """
    Set the meeting chairs.\
    See [meetbot plugin usage]({% link _usage/meetbot-plugin.md %})
    """
    if not is_meeting_running(trigger.sender):
        bot.say("There is no active meeting")
        return
    if not trigger.group(2):
        bot.say(
            "Who are the chairs? Try `{}chairs Alice Bob Cindy`".format(
                bot.config.core.help_prefix
            )
        )
        return
    if trigger.nick.lower() == meetings_dict[trigger.sender]["head"]:
        meetings_dict[trigger.sender]["chairs"] = trigger.group(2).lower().split(" ")
        chairs_readable = trigger.group(2).lower().replace(" ", ", ")
        log_plain("Meeting chairs are: " + chairs_readable, trigger.sender)
        log_html_listitem(
            "<span style='font-weight: bold'>Meeting chairs are:</span> %s"
            % chairs_readable,
            trigger.sender,
        )
        bot.say(formatting.bold("Meeting chairs are:") + " " + chairs_readable)
    else:
        bot.say("Only meeting head can set chairs")


# Log action item in the HTML log
@plugin.command("action")
@plugin.example(".action elad will develop a meetbot")
def meetingaction(bot, trigger):
    """
    Log an action in the meeting log.\
    See [meetbot plugin usage]({% link _usage/meetbot-plugin.md %})
    """
    if not is_meeting_running(trigger.sender):
        bot.say("There is no active meeting")
        return
    if not trigger.group(2):
        bot.say(
            "Try `{}action Bob will do something`".format(bot.config.core.help_prefix)
        )
        return
    if not is_chair(trigger.nick, trigger.sender):
        bot.say("Only meeting head or chairs can do that")
        return
    log_plain("ACTION: " + trigger.group(2), trigger.sender)
    log_html_listitem(
        "<span style='font-weight: bold'>Action: </span>" + trigger.group(2),
        trigger.sender,
    )
    meeting_actions[trigger.sender].append(trigger.group(2))
    bot.say(formatting.bold("ACTION:") + " " + trigger.group(2))


@plugin.command("listactions")
@plugin.example(".listactions")
def listactions(bot, trigger):
    """List all the actions in the meeting summary."""
    if not is_meeting_running(trigger.sender):
        bot.say("There is no active meeting")
        return
    for action in meeting_actions[trigger.sender]:
        bot.say(formatting.bold("ACTION:") + " " + action)


# Log agreed item in the HTML log
@plugin.command("agreed")
@plugin.example(".agreed Bowties are cool")
def meetingagreed(bot, trigger):
    """
    Log an agreement in the meeting log.\
    See [meetbot plugin usage]({% link _usage/meetbot-plugin.md %})
    """
    if not is_meeting_running(trigger.sender):
        bot.say("There is no active meeting")
        return
    if not trigger.group(2):
        bot.say("Try `{}agreed Bowties are cool`".format(bot.config.core.help_prefix))
        return
    if not is_chair(trigger.nick, trigger.sender):
        bot.say("Only meeting head or chairs can do that")
        return
    log_plain("AGREED: " + trigger.group(2), trigger.sender)
    log_html_listitem(
        "<span style='font-weight: bold'>Agreed: </span>" + trigger.group(2),
        trigger.sender,
    )
    bot.say(formatting.bold("AGREED:") + " " + trigger.group(2))


# Log link item in the HTML log
@plugin.command("link")
@plugin.example(".link http://example.com")
def meetinglink(bot, trigger):
    """
    Log a link in the meeing log.\
    See [meetbot plugin usage]({% link _usage/meetbot-plugin.md %})
    """
    if not is_meeting_running(trigger.sender):
        bot.say("There is no active meeting")
        return
    if not trigger.group(2):
        bot.say(
            "Try `{}link https://relevant-website.example/`".format(
                bot.config.core.help_prefix
            )
        )
        return
    if not is_chair(trigger.nick, trigger.sender):
        bot.say("Only meeting head or chairs can do that")
        return
    link = trigger.group(2)
    if not link.startswith("http"):
        link = "http://" + link
    try:
        title = find_title(link)
    except Exception:  # TODO: Be specific
        title = ""
    log_plain("LINK: %s [%s]" % (link, title), trigger.sender)
    log_html_listitem('<a href="%s">%s</a>' % (link, title), trigger.sender)
    bot.say(formatting.bold("LINK:") + " " + link)


# Log informational item in the HTML log
@plugin.command("info")
@plugin.example(".info all board members present")
def meetinginfo(bot, trigger):
    """
    Log an informational item in the meeting log.\
    See [meetbot plugin usage]({% link _usage/meetbot-plugin.md %})
    """
    if not is_meeting_running(trigger.sender):
        bot.say("There is no active meeting")
        return
    if not trigger.group(2):
        bot.say(
            "Try `{}info some informative thing`".format(bot.config.core.help_prefix)
        )
        return
    if not is_chair(trigger.nick, trigger.sender):
        bot.say("Only meeting head or chairs can do that")
        return
    log_plain("INFO: " + trigger.group(2), trigger.sender)
    log_html_listitem(trigger.group(2), trigger.sender)
    bot.say(formatting.bold("INFO:") + " " + trigger.group(2))


# called for every single message
# Will log to plain text only
@plugin.rule("(.*)")
@plugin.priority("low")
def log_meeting(bot, trigger):
    if not is_meeting_running(trigger.sender):
        return

    # Handle live prefix changes with cached regex
    if (
        "meetbot_prefix" not in bot.memory or
        bot.memory["meetbot_prefix"] != bot.config.core.prefix
    ):
        commands = [
            "startmeeting",
            "subject",
            "endmeeting",
            "chairs",
            "action",
            "listactions",
            "agreed",
            "link",
            "info",
            "comments",
        ]

        bot.memory["meetbot_command_regex"] = re.compile(
            "{}({})( |$)".format(bot.config.core.prefix, "|".join(commands))
        )
        bot.memory["meetbot_prefix"] = bot.config.core.prefix

    if bot.memory["meetbot_command_regex"].match(trigger):
        return
    log_plain("<" + trigger.nick + "> " + trigger, trigger.sender)


@plugin.command("comment")
@plugin.require_privmsg()
def take_comment(bot, trigger):
    """
    Log a comment, to be shown with other comments when a chair uses .comments.
    Intended to allow commentary from those outside the primary group of people
    in the meeting.

    Used in private message only, as `.comment <#channel> <comment to add>`

    See [meetbot plugin usage]({% link _usage/meetbot-plugin.md %})
    """
    if not trigger.group(4):  # <2 arguments were given
        bot.say(
            "Usage: {}comment <#channel> <comment to add>".format(
                bot.config.core.help_prefix
            )
        )
        return

    target, message = trigger.group(2).split(None, 1)
    target = tools.Identifier(target)
    if not is_meeting_running(target):
        bot.say("There is no active meeting in that channel.")
    else:
        meetings_dict[trigger.group(3)]["comments"].append((trigger.nick, message))
        bot.say(
            "Your comment has been recorded. It will be shown when the "
            "chairs tell me to show the comments."
        )
        bot.say(
            "A new comment has been recorded.", meetings_dict[trigger.group(3)]["head"]
        )


@plugin.command("comments")
def show_comments(bot, trigger):
    """
    Show the comments that have been logged for this meeting with .comment.

    See [meetbot plugin usage]({% link _usage/meetbot-plugin.md %})
    """
    if not is_meeting_running(trigger.sender):
        return
    if not is_chair(trigger.nick, trigger.sender):
        bot.say("Only meeting head or chairs can do that")
        return
    comments = meetings_dict[trigger.sender]["comments"]
    if comments:
        msg = "The following comments were made:"
        bot.say(msg)
        log_plain("<%s> %s" % (bot.nick, msg), trigger.sender)
        for comment in comments:
            msg = "<%s> %s" % comment
            bot.say(msg)
            log_plain("<%s> %s" % (bot.nick, msg), trigger.sender)
        meetings_dict[trigger.sender]["comments"] = []
    else:
        bot.say("No comments have been recorded")



# coding=utf-8
"""
unicode_info.py - Sopel Codepoints Plugin
Copyright 2013, Elsie Powell, embolalia.com
Copyright 2008, Sean B. Palmer, inamidst.com
Licensed under the Eiffel Forum License 2.

https://sopel.chat
"""
from __future__ import absolute_import, division, print_function, unicode_literals

import sys
import unicodedata

from sopel import plugin

if sys.version_info.major >= 3:
    # Note on unicode and str (required for py2 compatibility)
    # the `hex` function returns a `str`, both in py2 and py3
    # however, a `str` is a unicode string in py3, but a bytestring in py2
    # in order to prevent that, we encode the return from `hex` as `unicode`
    # and since this class does not exist anymore on py3, we create an alias
    # for `str` in py3
    unichr = chr
    unicode = str


def get_codepoint_name(char):
    """Retrieve the code point (and name, if possible) for a given character"""
    # Get the hex value for the code point, and drop the 0x from the front
    point = unicode(hex(ord(char)))[2:]

    # Make the hex 4 characters long with preceding 0s, and all upper case
    point = point.rjust(4, '0').upper()

    # get codepoint's name
    name = None
    try:
        name = unicodedata.name(char)
    except ValueError:
        pass

    return point, name


@plugin.command('u')
@plugin.example('.u ‽', 'U+203D INTERROBANG (‽)', user_help=True)
@plugin.example('.u 203D', 'U+203D INTERROBANG (‽)', user_help=True)
@plugin.output_prefix('[unicode] ')
def codepoint(bot, trigger):
    """Look up a Unicode character or a hexadecimal code point."""
    arg = trigger.group(2)
    if not arg:
        bot.reply('What code point do you want me to look up?')
        return plugin.NOLIMIT
    stripped = arg.strip()
    if len(stripped) > 0:
        arg = stripped
    if len(arg) > 1:
        if arg.startswith('U+'):
            arg = arg[2:]
        try:
            arg = unichr(int(arg, 16))
        except (ValueError, TypeError):
            bot.reply("That's not a valid code point.")
            return plugin.NOLIMIT

    point, name = get_codepoint_name(arg)
    if name is None:
        name = '(No name found)'

    template = 'U+%s %s (\xe2\x97\x8c%s)'
    if not unicodedata.combining(arg):
        template = 'U+%s %s (%s)'

    bot.say(template % (point, name, arg))



# coding=utf-8
"""Sopel's plugin handlers.

.. versionadded:: 7.0

Between a plugin (or "module") and Sopel's core, Plugin Handlers are used. It
is an interface (defined by the :class:`AbstractPluginHandler` abstract class),
that acts as a proxy between Sopel and the plugin, making a clear separation
between how the bot behaves and how the plugins work.

From the :class:`~sopel.bot.Sopel` class, a plugin must be:

* loaded, using :meth:`~AbstractPluginHandler.load`
* setup (if required), using :meth:`~AbstractPluginHandler.setup`
* and eventually registered using :meth:`~AbstractPluginHandler.register`

Each subclass of :class:`AbstractPluginHandler` must implement its methods in
order to be used in the application.

At the moment, three types of plugin are handled:

* :class:`PyModulePlugin`: manage plugins that can be imported as Python
  module from a Python package, i.e. where ``from package import name`` works
* :class:`PyFilePlugin`: manage plugins that are Python files on the filesystem
  or Python directory (with an ``__init__.py`` file inside), that cannot be
  directly imported and extra steps are necessary
* :class:`EntryPointPlugin`: manage plugins that are declared by a setuptools
  entry point; other than that, it behaves like a :class:`PyModulePlugin`

All expose the same interface and thereby abstract the internal implementation
away from the rest of the application.

.. important::

    This is all relatively new. Its usage and documentation is for Sopel core
    development and advanced developers. It is subject to rapid changes
    between versions without much (or any) warning.

    Do **not** build your plugin based on what is here, you do **not** need to.

"""
# Copyright 2019, Florian Strzelecki <florian.strzelecki@gmail.com>
#
# Licensed under the Eiffel Forum License 2.
from __future__ import absolute_import, division, print_function, unicode_literals

import imp
import importlib
import inspect
import itertools
import os

from sopel import loader
from . import exceptions

try:
    reload = importlib.reload
except AttributeError:
    # py2: no reload function
    # TODO: imp is deprecated, to be removed when py2 support is dropped
    reload = imp.reload


class AbstractPluginHandler(object):
    """Base class for plugin handlers.

    This abstract class defines the interface Sopel uses to
    configure, load, shutdown, etc. a Sopel plugin (or "module").

    It is through this interface that Sopel will interact with its plugins,
    whether internal (from ``sopel.modules``) or external (from the Python
    files in a directory, to ``sopel_modules.*`` subpackages).

    Sopel's loader will create a "Plugin Handler" for each plugin it finds, to
    which it then delegates loading the plugin, listing its functions
    (commands, jobs, etc.), configuring it, and running any required actions
    on shutdown (either upon exiting Sopel or unloading that plugin).
    """
    def load(self):
        """Load the plugin.

        This method must be called first, in order to setup, register, shutdown,
        or configure the plugin later.
        """
        raise NotImplementedError

    def reload(self):
        """Reload the plugin.

        This method can be called once the plugin is already loaded. It will
        take care of reloading the plugin from its source.
        """
        raise NotImplementedError

    def get_label(self):
        """Retrieve a display label for the plugin.

        :return: a human readable label for display purpose
        :rtype: str

        This method should, at least, return ``<module_name> plugin``.
        """
        raise NotImplementedError

    def get_meta_description(self):
        """Retrieve a meta description for the plugin.

        :return: meta description information
        :rtype: :class:`dict`

        The expected keys are:

        * name: a short name for the plugin
        * label: a descriptive label for the plugin
        * type: the plugin's type
        * source: the plugin's source
          (filesystem path, python import path, etc.)
        """
        raise NotImplementedError

    def is_loaded(self):
        """Tell if the plugin is loaded or not.

        :return: ``True`` if the plugin is loaded, ``False`` otherwise
        :rtype: bool

        This must return ``True`` if the :meth:`load` method has been called
        with success.
        """
        raise NotImplementedError

    def setup(self, bot):
        """Run the plugin's setup action.

        :param bot: instance of Sopel
        :type bot: :class:`sopel.bot.Sopel`
        """
        raise NotImplementedError

    def has_setup(self):
        """Tell if the plugin has a setup action.

        :return: ``True`` if the plugin has a setup, ``False`` otherwise
        :rtype: bool
        """
        raise NotImplementedError

    def register(self, bot):
        """Register the plugin with the ``bot``.

        :param bot: instance of Sopel
        :type bot: :class:`sopel.bot.Sopel`
        """
        raise NotImplementedError

    def unregister(self, bot):
        """Unregister the plugin from the ``bot``.

        :param bot: instance of Sopel
        :type bot: :class:`sopel.bot.Sopel`
        """
        raise NotImplementedError

    def shutdown(self, bot):
        """Run the plugin's shutdown action.

        :param bot: instance of Sopel
        :type bot: :class:`sopel.bot.Sopel`
        """
        raise NotImplementedError

    def has_shutdown(self):
        """Tell if the plugin has a shutdown action.

        :return: ``True`` if the plugin has a ``shutdown`` action, ``False``
                 otherwise
        :rtype: bool
        """
        raise NotImplementedError

    def configure(self, settings):
        """Configure Sopel's ``settings`` for this plugin.

        :param settings: Sopel's configuration
        :type settings: :class:`sopel.config.Config`

        This method will be called by Sopel's configuration wizard.
        """
        raise NotImplementedError

    def has_configure(self):
        """Tell if the plugin has a configure action.

        :return: ``True`` if the plugin has a ``configure`` action, ``False``
                 otherwise
        :rtype: bool
        """
        raise NotImplementedError


class PyModulePlugin(AbstractPluginHandler):
    """Sopel plugin loaded from a Python module or package.

    A :class:`PyModulePlugin` represents a Sopel plugin that is a Python
    module (or package) that can be imported directly.

    This::

        >>> import sys
        >>> from sopel.plugins.handlers import PyModulePlugin
        >>> plugin = PyModulePlugin('xkcd', 'sopel.modules')
        >>> plugin.module_name
        'sopel.modules.xkcd'
        >>> plugin.load()
        >>> plugin.module_name in sys.modules
        True

    Is the same as this::

        >>> import sys
        >>> from sopel.modules import xkcd
        >>> 'sopel.modules.xkcd' in sys.modules
        True

    """

    PLUGIN_TYPE = 'python-module'
    """The plugin's type.

    Metadata for the plugin; this should be considered to be a constant and
    should not be modified at runtime.
    """

    def __init__(self, name, package=None):
        self.name = name
        self.package = package
        if package:
            self.module_name = self.package + '.' + self.name
        else:
            self.module_name = name

        self._module = None

    def get_label(self):
        """Retrieve a display label for the plugin.

        :return: a human readable label for display purpose
        :rtype: str

        By default, this is ``<name> plugin``. If the plugin's module has a
        docstring, its first line is used as the plugin's label.
        """
        default_label = '%s plugin' % self.name
        module_doc = getattr(self._module, '__doc__', None)

        if not self.is_loaded() or not module_doc:
            return default_label

        lines = inspect.cleandoc(module_doc).splitlines()
        return default_label if not lines else lines[0]

    def get_meta_description(self):
        """Retrieve a meta description for the plugin.

        :return: meta description information
        :rtype: :class:`dict`

        The keys are:

        * name: the plugin's name
        * label: see :meth:`~sopel.plugins.handlers.PyModulePlugin.get_label`
        * type: see :attr:`PLUGIN_TYPE`
        * source: the name of the plugin's module

        Example::

            {
                'name': 'example',
                'type: 'python-module',
                'label: 'example plugin',
                'source': 'sopel_modules.example',
            }

        """
        return {
            'label': self.get_label(),
            'type': self.PLUGIN_TYPE,
            'name': self.name,
            'source': self.module_name,
        }

    def load(self):
        """Load the plugin's module using :func:`importlib.import_module`.

        This method assumes the module is available through ``sys.path``.
        """
        self._module = importlib.import_module(self.module_name)

    def reload(self):
        """Reload the plugin's module using :func:`importlib.reload`.

        This method assumes the plugin is already loaded.
        """
        self._module = reload(self._module)

    def is_loaded(self):
        return self._module is not None

    def setup(self, bot):
        if self.has_setup():
            self._module.setup(bot)

    def has_setup(self):
        """Tell if the plugin has a setup action.

        :return: ``True`` if the plugin has a setup, ``False`` otherwise
        :rtype: bool

        The plugin has a setup action if its module has a ``setup`` attribute.
        This attribute is expected to be a callable.
        """
        return hasattr(self._module, 'setup')

    def register(self, bot):
        relevant_parts = loader.clean_module(self._module, bot.config)
        for part in itertools.chain(*relevant_parts):
            # annotate all callables in relevant_parts with `plugin_name`
            # attribute to make per-channel config work; see #1839
            setattr(part, 'plugin_name', self.name)
        bot.add_plugin(self, *relevant_parts)

    def unregister(self, bot):
        relevant_parts = loader.clean_module(self._module, bot.config)
        bot.remove_plugin(self, *relevant_parts)

    def shutdown(self, bot):
        if self.has_shutdown():
            self._module.shutdown(bot)

    def has_shutdown(self):
        """Tell if the plugin has a shutdown action.

        :return: ``True`` if the plugin has a ``shutdown`` action, ``False``
                 otherwise
        :rtype: bool

        The plugin has a shutdown action if its module has a ``shutdown``
        attribute. This attribute is expected to be a callable.
        """
        return hasattr(self._module, 'shutdown')

    def configure(self, settings):
        if self.has_configure():
            self._module.configure(settings)

    def has_configure(self):
        """Tell if the plugin has a configure action.

        :return: ``True`` if the plugin has a ``configure`` action, ``False``
                 otherwise
        :rtype: bool

        The plugin has a configure action if its module has a ``configure``
        attribute. This attribute is expected to be a callable.
        """
        return hasattr(self._module, 'configure')


class PyFilePlugin(PyModulePlugin):
    """Sopel plugin loaded from the filesystem outside of the Python path.

    This plugin handler can be used to load a Sopel plugin from the
    filesystem, either a Python ``.py`` file or a directory containing an
    ``__init__.py`` file, and behaves like a :class:`PyModulePlugin`::

        >>> from sopel.plugins.handlers import PyFilePlugin
        >>> plugin = PyFilePlugin('/home/sopel/.sopel/modules/custom.py')
        >>> plugin.load()
        >>> plugin.name
        'custom'

    In this example, the plugin ``custom`` is loaded from its filename despite
    not being in the Python path.
    """

    PLUGIN_TYPE = 'python-file'
    """The plugin's type.

    Metadata for the plugin; this should be considered to be a constant and
    should not be modified at runtime.
    """

    def __init__(self, filename):
        good_file = (
            os.path.isfile(filename) and
            filename.endswith('.py') and not filename.startswith('_')
        )
        good_dir = (
            os.path.isdir(filename) and
            os.path.isfile(os.path.join(filename, '__init__.py'))
        )

        if good_file:
            name = os.path.basename(filename)[:-3]
            module_type = imp.PY_SOURCE
        elif good_dir:
            name = os.path.basename(filename)
            module_type = imp.PKG_DIRECTORY
        else:
            raise exceptions.PluginError('Invalid Sopel plugin: %s' % filename)

        self.filename = filename
        self.path = filename
        self.module_type = module_type

        super(PyFilePlugin, self).__init__(name)

    def _load(self):
        # The current implementation uses `imp.load_module` to perform the
        # load action, which also reloads the module. However, `imp` is
        # deprecated in Python 3, so that might need to be changed when the
        # support for Python 2 is dropped.
        #
        # However, the solution for Python 3 is non-trivial, since the
        # `importlib` built-in module does not have a similar function,
        # therefore requires to dive into its public internals
        # (``importlib.machinery`` and ``importlib.util``).
        #
        # All of that is doable, but represents a lot of work. As long as
        # Python 2 is supported, we can keep it for now.
        #
        # TODO: switch to ``importlib`` when Python2 support is dropped.
        if self.module_type == imp.PY_SOURCE:
            with open(self.path) as mod:
                description = ('.py', 'U', self.module_type)
                mod = imp.load_module(self.name, mod, self.path, description)
        elif self.module_type == imp.PKG_DIRECTORY:
            description = ('', '', self.module_type)
            mod = imp.load_module(self.name, None, self.path, description)
        else:
            raise TypeError('Unsupported module type')

        return mod

    def get_meta_description(self):
        """Retrieve a meta description for the plugin.

        :return: meta description information
        :rtype: :class:`dict`

        This returns the same keys as
        :meth:`PyModulePlugin.get_meta_description`; the ``source`` key is
        modified to contain the source file's path instead of its Python module
        dotted path::

            {
                'name': 'example',
                'type: 'python-file',
                'label: 'example plugin',
                'source': '/home/username/.sopel/plugins/example.py',
            }

        """
        data = super(PyFilePlugin, self).get_meta_description()
        data.update({
            'source': self.path,
        })
        return data

    def load(self):
        self._module = self._load()

    def reload(self):
        """Reload the plugin.

        Unlike :class:`PyModulePlugin`, it is not possible to use the
        ``reload`` function (either from `imp` or `importlib`), because the
        module might not be available through ``sys.path``.
        """
        self._module = self._load()


class EntryPointPlugin(PyModulePlugin):
    """Sopel plugin loaded from a ``setuptools`` entry point.

    :param entry_point: a ``setuptools`` entry point object

    This handler loads a Sopel plugin exposed by a ``setuptools`` entry point.
    It expects to be able to load a module object from the entry point, and to
    work as a :class:`~.PyModulePlugin` from that module.

    By default, Sopel uses the entry point ``sopel.plugins``. To use that for
    their plugin, developers must define an entry point either in their
    ``setup.py`` file or their ``setup.cfg`` file::

        # in setup.py file
        setup(
            name='my_plugin',
            version='1.0',
            entry_points={
                'sopel.plugins': [
                    'custom = my_plugin.path.to.plugin',
                ],
            }
        )

    And this plugin can be loaded with::

        >>> from pkg_resources import iter_entry_points
        >>> from sopel.plugins.handlers import EntryPointPlugin
        >>> plugin = [
        ...     EntryPointPlugin(ep)
        ...     for ep in iter_entry_points('sopel.plugins', 'custom')
        ... ][0]
        >>> plugin.load()
        >>> plugin.name
        'custom'

    In this example, the plugin ``custom`` is loaded from an entry point.
    Unlike the :class:`~.PyModulePlugin`, the name is not derived from the
    actual Python module, but from its entry point's name.

    .. seealso::

        Sopel uses the :func:`~sopel.plugins.find_entry_point_plugins` function
        internally to search entry points.

        Entry point is a `standard feature of setuptools`__ for Python, used
        by other applications (like ``pytest``) for their plugins.

        .. __: https://setuptools.readthedocs.io/en/stable/setuptools.html#dynamic-discovery-of-services-and-plugins

    """

    PLUGIN_TYPE = 'setup-entrypoint'
    """The plugin's type.

    Metadata for the plugin; this should be considered to be a constant and
    should not be modified at runtime.
    """

    def __init__(self, entry_point):
        self.entry_point = entry_point
        super(EntryPointPlugin, self).__init__(entry_point.name)

    def load(self):
        self._module = self.entry_point.load()

    def get_meta_description(self):
        """Retrieve a meta description for the plugin.

        :return: meta description information
        :rtype: :class:`dict`

        This returns the same keys as
        :meth:`PyModulePlugin.get_meta_description`; the ``source`` key is
        modified to contain the setuptools entry point::

            {
                'name': 'example',
                'type: 'setup-entrypoint',
                'label: 'example plugin',
                'source': 'example = my_plugin.example',
            }

        """
        data = super(EntryPointPlugin, self).get_meta_description()
        data.update({
            'source': str(self.entry_point),
        })
        return data



# coding=utf-8
"""
ip.py - Sopel GeoIP Lookup Plugin
Copyright 2011, Dimitri Molenaars, TyRope.nl,
Copyright © 2013, Elad Alfassa <elad@fedoraproject.org>
Licensed under the Eiffel Forum License 2.

https://sopel.chat
"""

from __future__ import absolute_import, division, print_function, unicode_literals

import logging
import os
import socket
import tarfile

import geoip2.database

from sopel import plugin
from sopel.config import types
from sopel.tools import web

urlretrieve = None
try:
    from urllib import urlretrieve
except ImportError:
    try:
        # urlretrieve has been put under urllib.request in Python 3.
        # It's also deprecated so this should probably be replaced with
        # urllib2.
        from urllib.request import urlretrieve
    except ImportError:
        pass


LOGGER = logging.getLogger(__name__)


class GeoipSection(types.StaticSection):
    GeoIP_db_path = types.FilenameAttribute('GeoIP_db_path', directory=True)
    """Path of the directory containing the GeoIP database files.

    If the given value is not an absolute path, it will be interpreted relative
    to the directory containing the config file with which Sopel was started.
    """


def configure(config):
    """
    | name | example | purpose |
    | ---- | ------- | ------- |
    | GeoIP\\_db\\_path | /home/sopel/GeoIP/ | Path to the GeoIP database files |
    """
    config.define_section('ip', GeoipSection)
    config.ip.configure_setting('GeoIP_db_path',
                                'Path of the GeoIP db files')


def setup(bot):
    bot.config.define_section('ip', GeoipSection)


def _decompress(source, target, delete_after_decompression=True):
    """Decompress just the database from the archive"""
    # https://stackoverflow.com/a/16452962
    tar = tarfile.open(source)
    for member in tar.getmembers():
        if ".mmdb" in member.name:
            member.name = os.path.basename(member.name)
            tar.extract(member, target)
    if delete_after_decompression:
        os.remove(source)


def _find_geoip_db(bot):
    """Find the GeoIP database"""
    config = bot.config
    if config.ip.GeoIP_db_path:
        cities_db = os.path.join(config.ip.GeoIP_db_path, 'GeoLite2-City.mmdb')
        ipasnum_db = os.path.join(config.ip.GeoIP_db_path, 'GeoLite2-ASN.mmdb')
        if (os.path.isfile(cities_db) and os.path.isfile(ipasnum_db)):
            return config.ip.GeoIP_db_path
        else:
            LOGGER.warning(
                'GeoIP path configured but DB not found in configured path')

    if (os.path.isfile(os.path.join(config.core.homedir, 'GeoLite2-City.mmdb')) and
            os.path.isfile(os.path.join(config.core.homedir, 'GeoLite2-ASN.mmdb'))):
        return config.core.homedir
    elif (os.path.isfile(os.path.join('/usr/share/GeoIP', 'GeoLite2-City.mmdb')) and
            os.path.isfile(os.path.join('/usr/share/GeoIP', 'GeoLite2-ASN.mmdb'))):
        return '/usr/share/GeoIP'
    elif urlretrieve:
        LOGGER.info('Downloading GeoIP database')
        bot.say('Downloading GeoIP database, please wait...')

        common_params = {'license_key': 'JXBEmLjOzislFnh4', 'suffix': 'tar.gz'}
        base_url = 'https://download.maxmind.com/app/geoip_download'
        geolite_urls = []

        for edition in ['ASN', 'City']:
            geolite_urls.append(
                '{base}?{params}'.format(
                    base=base_url,
                    params=web.urlencode(dict(common_params, **{'edition_id': 'GeoLite2-%s' % edition})),
                )
            )

        for url in geolite_urls:
            LOGGER.debug('GeoIP Source URL: %s', url)
            full_path = os.path.join(config.core.homedir, url.split("/")[-1])
            urlretrieve(url, full_path)
            _decompress(full_path, config.core.homedir)
        return bot.config.core.homedir
    else:
        return False


@plugin.command('iplookup', 'ip')
@plugin.example(
    '.ip 8.8.8.8',
    r'Hostname: \S*dns\S*\.google\S*( \| .+?: .+?)+ \| ISP: AS15169 \S+',
    re=True,
    ignore='Downloading GeoIP database, please wait...',
    online=True)
@plugin.output_prefix('[IP/Host Lookup] ')
def ip(bot, trigger):
    """IP Lookup tool"""
    # Check if there is input at all
    if not trigger.group(2):
        bot.reply("No search term.")
        return

    parts = []
    # Check whether the input is an IP or hostmask or a nickname
    decide = ['.', ':']
    if any(x in trigger.group(2) for x in decide):
        # It's an IP/hostname!
        query = trigger.group(2).strip()
    else:
        # Need to get the host for the username
        username = trigger.group(2).strip().lower()
        user_in_botdb = bot.users.get(username)
        if user_in_botdb is not None:
            query = user_in_botdb.host

            # Sanity check - sometimes user information isn't populated yet
            if query is None:
                bot.reply("I don't know that user's host.")
                return
        else:
            bot.reply("I\'m not aware of this user.")
            return

    db_path = _find_geoip_db(bot)
    if db_path is False:
        LOGGER.error('Can\'t find (or download) usable GeoIP database.')
        bot.reply('Sorry, I don\'t have a GeoIP database to use for this lookup.')
        return

    if ':' in query:
        try:
            socket.inet_pton(socket.AF_INET6, query)
        except (OSError, socket.error):  # Python 2/3 compatibility
            bot.reply("Unable to resolve IP/Hostname")
            return
    elif '.' in query:
        try:
            socket.inet_pton(socket.AF_INET, query)
        except (socket.error, socket.herror):
            try:
                query = socket.getaddrinfo(query, None)[0][4][0]
            except socket.gaierror:
                bot.reply("Unable to resolve IP/Hostname")
                return
    else:
        bot.reply("Unable to resolve IP/Hostname")
        return

    city = geoip2.database.Reader(os.path.join(db_path, 'GeoLite2-City.mmdb'))
    asn = geoip2.database.Reader(os.path.join(db_path, 'GeoLite2-ASN.mmdb'))
    host = socket.getfqdn(query)
    try:
        city_response = city.city(query)
        asn_response = asn.asn(query)
    except geoip2.errors.AddressNotFoundError:
        bot.reply("The address is not in the database.")
        return

    parts.append("Hostname: %s" % host)
    try:
        parts.append("Location: %s" % city_response.country.name)
    except AttributeError:
        parts.append('Location: Unknown')

    region = city_response.subdivisions.most_specific.name
    if region:
        parts.append("Region: %s" % region)

    city = city_response.city.name
    if city:
        parts.append("City: %s" % city)

    isp = "ISP: AS" + str(asn_response.autonomous_system_number) + \
          " " + asn_response.autonomous_system_organization
    parts.append(isp)

    bot.say(' | '.join(parts))



# coding=utf-8
"""
tld.py - Sopel TLD Plugin
Copyright 2009-10, Michael Yanovich, yanovich.net
Copyright 2020, dgw, technobabbl.es
Licensed under the Eiffel Forum License 2.

https://sopel.chat
"""
from __future__ import absolute_import, division, print_function, unicode_literals

from datetime import datetime
from encodings import idna
import logging
import re
import sys

import pytz
import requests

from sopel import formatting, plugin, tools

if sys.version_info.major >= 3:
    unicode = str
    from html.parser import HTMLParser
else:
    from HTMLParser import HTMLParser


LOGGER = logging.getLogger(__name__)


def _strptime_as_utc(time_string):
    return datetime.strptime(
        time_string, DATE_FORMAT).replace(tzinfo=pytz.utc)


DATE_FORMAT = '%Y-%m-%d %H:%M:%S'
DEFAULT_CACHE_TIME = '2000-01-01 00:00:00'
DEFAULT_CACHE_DATETIME = _strptime_as_utc(DEFAULT_CACHE_TIME)

IANA_LIST_URI = 'https://data.iana.org/TLD/tlds-alpha-by-domain.txt'
WIKI_PAGE_NAMES = [
    'List_of_Internet_top-level_domains',
    'Country_code_top-level_domain',
]
r_tld = re.compile(r'^\.(\S+)')
r_idn = re.compile(r'^(xn--[A-Za-z0-9]+)')


def setup(bot):
    bot.memory['tld_list_cache'] = bot.db.get_plugin_value(
        'tld', 'tld_list_cache', [])
    bot.memory['tld_list_cache_updated'] = bot.db.get_plugin_value(
        'tld', 'tld_list_cache_updated', DEFAULT_CACHE_TIME)
    bot.memory['tld_data_cache'] = bot.db.get_plugin_value(
        'tld', 'tld_data_cache', {})
    bot.memory['tld_data_cache_updated'] = bot.db.get_plugin_value(
        'tld', 'tld_data_cache_updated', DEFAULT_CACHE_TIME)

    # Restore datetime objects from string format. If parsing fails, consider
    # that cache obsolete so it will be updated at the next age check.
    try:
        bot.memory['tld_list_cache_updated'] = _strptime_as_utc(
            bot.memory['tld_list_cache_updated'])
    except ValueError:
        bot.memory['tld_list_cache_updated'] = DEFAULT_CACHE_DATETIME
    try:
        bot.memory['tld_data_cache_updated'] = _strptime_as_utc(
            bot.memory['tld_data_cache_updated'])
    except ValueError:
        bot.memory['tld_data_cache_updated'] = DEFAULT_CACHE_DATETIME


def shutdown(bot):
    if bot.memory.get('tld_list_cache'):
        bot.db.set_plugin_value(
            'tld', 'tld_list_cache', bot.memory['tld_list_cache'])
        bot.db.set_plugin_value(
            'tld', 'tld_list_cache_updated',
            bot.memory['tld_list_cache_updated'].strftime(DATE_FORMAT))
    if bot.memory.get('tld_data_cache'):
        bot.db.set_plugin_value(
            'tld', 'tld_data_cache', bot.memory['tld_data_cache'])
        bot.db.set_plugin_value(
            'tld', 'tld_data_cache_updated',
            bot.memory['tld_data_cache_updated'].strftime(DATE_FORMAT))

    for key in [
        'tld_list_cache',
        'tld_list_cache_updated',
        'tld_data_cache',
        'tld_data_cache_updated',
    ]:
        try:
            del bot.memory[key]
        except KeyError:
            pass


class WikipediaTLDListParser(HTMLParser):
    def __init__(self):
        HTMLParser.__init__(self)
        self.in_cell = False
        self.skipping = True
        self.current_row = []
        self.current_cell = ''
        self.rows = []
        self.tables = []
        self.finished = False

    def handle_starttag(self, tag, attrs):
        if tag == 'td' or tag == 'th':
            self.in_cell = True
        elif tag == 'sup' and self.in_cell:
            # ignore superscripts; they're almost exclusively footnotes
            self.skipping = True
        elif tag == 'table':
            for name, value in attrs:
                if name == 'class' and 'wikitable' in value:
                    self.skipping = False
        elif tag in ['b', 'strong'] and self.in_cell:
            self.current_cell += '<[bold]>'
        elif tag in ['i', 'em'] and self.in_cell:
            self.current_cell += '<[italic]>'

    def handle_endtag(self, tag):
        if tag == 'td' or tag == 'th':
            self.in_cell = False
            if not self.skipping:
                cell = self.current_cell.strip()
                # Python's built-in `strip()` method for strings will remove
                # some control codes we want to keep as IRC formatting. So the
                # parser inserts placeholders, and now it's time to replace
                # them with the real control codes.
                for placeholder in [
                    ('<[bold]>', formatting.CONTROL_BOLD),
                    ('<[italic]>', formatting.CONTROL_ITALIC),
                ]:
                    cell = cell.replace(*placeholder)
                self.current_row.append(cell)
            self.current_cell = ''
        elif tag == 'tr':
            if not self.skipping:
                self.rows.append(tuple(self.current_row))
            self.current_row = []
        elif tag == 'table':
            if not self.skipping:
                self.tables.append(self.rows)
            self.rows = []
            self.skipping = True
            self.in_cell = False
        elif tag == 'sup' and self.in_cell:
            self.skipping = False
        elif tag in ['b', 'strong'] and self.in_cell:
            self.current_cell += '<[bold]>'
        elif tag in ['i', 'em'] and self.in_cell:
            self.current_cell += '<[italic]>'

    def handle_data(self, data):
        if self.in_cell and not self.skipping:
            self.current_cell += data

    def get_processed_data(self):
        LOGGER.debug("Processed TLD data requested.")
        if self.finished:
            LOGGER.debug("Returning stored previously-processed data.")
            return self.tables

        LOGGER.debug("Ensuring all buffered data has been parsed.")
        self.close()

        LOGGER.debug("Processing tables.")
        tld_list = {}
        for table in self.tables:
            headings = table[0]
            for row in table[1:]:
                key = None
                idn_key = None
                for cell in row:
                    tld = r_tld.match(cell)
                    if tld and not key:
                        key = tld.group(1).lower()
                    idn = r_idn.match(cell)
                    if idn and not idn_key:
                        idn_key = idn.group(1).lower()
                if not any([key, idn_key]):
                    LOGGER.warning(
                        "Skipping row %s; could not find string to use as lookup key.",
                        str(row),
                    )
                    continue

                # Some cleanup happens directly in the dict comprehension here.
                # Empty values (actually falsy, but only empty strings are possible)
                # and values consisting only of a dash (indicating the absence of
                # information or restrictions) get left out of the final data.
                # When the data is presented later, these empty fields are just
                # clutter taking up limited space in the IRC line.
                zipped = {
                    field: value
                    for field, value
                    in dict(zip(headings, row)).items()
                    if value and value != '—'
                }
                if key:
                    tld_list[key] = zipped
                if idn_key:
                    tld_list[idn_key] = zipped

        LOGGER.debug("Finished processing TLD data; returning it.")
        self.tables = tld_list
        self.finished = True
        return self.tables


def _update_tld_data(bot, which, force=False):
    if which == 'list':
        then = bot.memory['tld_list_cache_updated']
    elif which == 'data':
        then = bot.memory['tld_data_cache_updated']
    else:
        LOGGER.error("Asked to update unknown cache type '%s'.", which)
        return

    now = datetime.now(tz=pytz.utc)
    since = now - then
    if not force and since.days < 7:
        LOGGER.debug(
            "Skipping TLD %s cache update; the cache is only %s old.",
            which,
            '1 day' if since.days == 1 else ('%d days' % since.days),
        )
        return

    if which == 'list':
        try:
            tld_list = requests.get(IANA_LIST_URI).text
        except requests.exceptions.RequestException:
            # Probably a transient error; log it and continue life
            LOGGER.warning(
                "Error fetching IANA TLD list; will try again later.",
                exc_info=True)
            return

        tld_list = [
            line.lower()
            for line in tld_list.splitlines()
            if not line.startswith('#')
        ]

        bot.memory['tld_list_cache'] = tld_list
        bot.memory['tld_list_cache_updated'] = now
    elif which == 'data':
        data_pages = []
        for title in WIKI_PAGE_NAMES:
            try:
                # https://www.mediawiki.org/wiki/Special:MyLanguage/API:Get_the_contents_of_a_page
                tld_response = requests.get(
                    "https://en.wikipedia.org/w/api.php",
                    params={
                        "action": "parse",
                        "format": "json",
                        "prop": "text",
                        "utf8": 1,
                        "formatversion": 2,
                        "page": title,
                    },
                ).json()
                data_pages.append(tld_response["parse"]["text"])
            # py <3.5 needs ValueError instead of more specific json.decoder.JSONDecodeError
            except (requests.exceptions.RequestException, ValueError, KeyError):
                # Log error and continue life; it'll be fine
                LOGGER.warning(
                    'Error fetching TLD data from "%s" on Wikipedia; will try again later.',
                    title, exc_info=True)
                return

        parser = WikipediaTLDListParser()
        for page in data_pages:
            parser.feed(page)

        tld_data = parser.get_processed_data()

        bot.memory['tld_data_cache'] = tld_data
        bot.memory['tld_data_cache_updated'] = now

    LOGGER.debug("Updated TLD %s cache.", which)


@plugin.interval(60 * 60)
def update_caches(bot, force=False):
    _update_tld_data(bot, 'list', force)
    _update_tld_data(bot, 'data', force)


@plugin.command('tld')
@plugin.example('.tld ru')
@plugin.output_prefix('[tld] ')
def gettld(bot, trigger):
    """Show information about the given Top Level Domain."""
    tld = trigger.group(2)
    if not tld:
        bot.reply("You must provide a top-level domain to search.")
        return  # Stop if no tld argument is provided
    tld = tld.strip('.').lower()

    if not bot.memory.get('tld_list_cache'):
        _update_tld_data(bot, 'list')
    tld_list = bot.memory['tld_list_cache']

    if not any([
        name in tld_list
        for name
        in [tld, idna.ToASCII(tld).decode('utf-8')]
    ]):
        bot.reply(
            "The top-level domain '{}' is not in IANA's list of valid TLDs."
            .format(tld))
        return

    if not bot.memory.get('tld_data_cache'):
        _update_tld_data(bot, 'data')
    tld_data = bot.memory['tld_data_cache']

    record = tld_data.get(tld, None)
    if not record:
        bot.say(
            "The top-level domain '{}' exists, "
            "but no details about it could be found."
            .format(tld)
        )
        return

    # Get the current order of available data fields
    fields = list(record.keys())
    # This trick moves matching keys to the end of the list
    fields.sort(key=lambda s: s.startswith('Notes') or s.startswith('Comments'))

    items = []
    for field in fields:
        value = record[field]
        if value:
            items.append('{}: {}'.format(field, value))

    message = ' | '.join(items)

    bot.say(message, truncation=' […]')


@plugin.command('tldcache')
@plugin.example('.tldcache clear', user_help=True)
@plugin.example('.tldcache update', user_help=True)
@plugin.example('.tldcache', user_help=True)
@plugin.require_admin
def tld_cache_command(bot, trigger):
    subcommand = trigger.group(2)
    if subcommand is None:
        times = {}
        for kind in ['list', 'data']:
            time = bot.memory.get('tld_' + kind + '_cache_updated')

            if time is None or time == DEFAULT_CACHE_DATETIME:
                times[kind] = '(not cached)'
                continue

            times[kind] = tools.time.format_time(
                db=bot.db,
                config=bot.config,
                nick=trigger.nick,
                channel=trigger.sender if not trigger.sender.is_nick() else None,
                time=time,
            )

        bot.reply(
            "IANA list updated at {}; Wikipedia data last fetched at {}."
            .format(times['list'], times['data']))
        return

    subcommand = subcommand.lower()
    if subcommand == 'update':
        bot.reply("Requesting updated IANA list and Wikipedia data.")
        update_caches(bot, force=True)
    elif subcommand == 'clear':
        bot.db.delete_plugin_value('tld', 'tld_data_cache')
        bot.memory['tld_data_cache'] = {}
        bot.db.delete_plugin_value('tld', 'tld_list_cache')
        bot.memory['tld_list_cache'] = []
        bot.db.delete_plugin_value('tld', 'tld_data_cache_updated')
        bot.memory['tld_data_cache_updated'] = DEFAULT_CACHE_DATETIME
        bot.db.delete_plugin_value('tld', 'tld_list_cache_updated')
        bot.memory['tld_list_cache_updated'] = DEFAULT_CACHE_DATETIME

        bot.reply("Cleared all cached TLD data.")
    else:
        bot.reply(
            "Unknown subcommand '{}'; recognized values are 'update' and 'clear'."
            .format(subcommand)
        )



# coding=utf-8
"""Sopel's plugin jobs management.

.. versionadded:: 7.1

.. important::

    This is all fresh and new. Its usage and documentation is for Sopel core
    development and advanced developers. It is subject to rapid changes
    between versions without much (or any) warning.

    Do **not** build your plugin based on what is here, you do **not** need to.

"""
# Copyright 2020, Florian Strzelecki <florian.strzelecki@gmail.com>
#
# Licensed under the Eiffel Forum License 2.
from __future__ import absolute_import, division, print_function, unicode_literals

import itertools
import logging

from sopel import tools
from sopel.tools import jobs

LOGGER = logging.getLogger(__name__)


class Scheduler(jobs.Scheduler):
    """Plugin job scheduler.

    :param manager: bot instance passed to jobs as argument
    :type manager: :class:`sopel.bot.Sopel`

    Scheduler that stores plugin jobs and behaves like its
    :class:`parent class <sopel.tools.jobs.Scheduler>`.

    .. versionadded:: 7.1

    .. note::

        This class is a specific implementation of the scheduler, made to store
        jobs by their plugins and be used by the bot (its ``manager``).
        It follows a similar interface as the
        :class:`plugin rules manager <sopel.plugins.rules.Manager>`.

    .. important::

        This is an internal tool used by Sopel to manage its jobs. To register
        a job, plugin authors should use :func:`sopel.plugin.interval`.

    """
    def __init__(self, manager):
        super(Scheduler, self).__init__(manager)
        self._jobs = tools.SopelMemoryWithDefault(list)

    def register(self, job):
        with self._mutex:
            self._jobs[job.get_plugin_name()].append(job)
        LOGGER.debug('Job registered: %s', str(job))

    def unregister_plugin(self, plugin_name):
        """Unregister all the jobs from a plugin.

        :param str plugin_name: the name of the plugin to remove
        :return: the number of jobs unregistered for this plugin
        :rtype: int

        All jobs of that plugin will be removed from the scheduler.

        This method is thread safe. However, it won't cancel or stop any
        currently running jobs.
        """
        unregistered_jobs = 0
        with self._mutex:
            jobs_count = len(self._jobs[plugin_name])
            del self._jobs[plugin_name]
            unregistered_jobs = unregistered_jobs + jobs_count

        LOGGER.debug(
            '[%s] Successfully unregistered %d jobs',
            plugin_name,
            unregistered_jobs)

        return unregistered_jobs

    def clear_jobs(self):
        with self._mutex:
            self._jobs = tools.SopelMemoryWithDefault(list)

        LOGGER.debug('Successfully unregistered all jobs')

    def remove_callable_job(self, callable):
        plugin_name = getattr(callable, 'plugin_name', None)
        if not self._jobs[plugin_name]:
            return

        with self._mutex:
            self._jobs[plugin_name] = [
                job for job in self._jobs[plugin_name]
                if job._handler != callable
            ]

    def _get_ready_jobs(self, now):
        with self._mutex:
            jobs = [
                job for job in itertools.chain(*self._jobs.values())
                if job.is_ready_to_run(now)
            ]

        return jobs



# coding=utf-8
"""
rand.py - Rand Plugin
Copyright 2013, Ari Koivula, <ari@koivu.la>
Licensed under the Eiffel Forum License 2.

https://sopel.chat
"""
from __future__ import absolute_import, division, print_function, unicode_literals

import random
import sys

from sopel import plugin


@plugin.command('rand')
@plugin.example('.rand 2', r'random\(0, 2\): (0|1|2)', re=True, repeat=10)
@plugin.example('.rand -1 -1', 'random(-1, -1): -1')
@plugin.example('.rand', r'random\(0, \d+\): \d+', re=True)
@plugin.example('.rand 99 10', r'random\(10, 99\): \d\d', re=True, repeat=10)
@plugin.example('.rand 10 99', r'random\(10, 99\): \d\d', re=True, repeat=10)
@plugin.output_prefix('[rand] ')
def rand(bot, trigger):
    """Replies with a random number between first and second argument."""
    arg1 = trigger.group(3)
    arg2 = trigger.group(4)

    try:
        if arg2 is not None:
            low = int(arg1)
            high = int(arg2)
        elif arg1 is not None:
            low = 0
            high = int(arg1)
        else:
            low = 0
            high = sys.maxsize
    except (ValueError, TypeError):
        return bot.reply("Arguments must be integers.")

    if low > high:
        low, high = high, low

    number = random.randint(low, high)
    bot.say("random(%d, %d): %d" % (low, high, number))



# coding=utf-8
"""This module provided tools that helped to write tests.

.. deprecated:: 7.1

    This module will be **removed in Sopel 8**.

    It formerly contained mock classes for the bot, its wrapper, and its config
    object. As the module is deprecated, so are they, and they will be removed
    as well.

    New code should use the :mod:`pytest plugin <sopel.tests.pytest_plugin>`
    for Sopel; or should take advantage of the :mod:`~sopel.tests.mocks` and
    :mod:`~sopel.tests.factories` modules, both added in Sopel 7.0.

"""
# Copyright 2013, Ari Koivula, <ari@koivu.la>
# Copyright 2019, Florian Strzelecki <florian.strzelecki@gmail.com>
# Licensed under the Eiffel Forum License 2.
from __future__ import absolute_import, division, print_function, unicode_literals

import os
import re
import sys
import tempfile

try:
    import ConfigParser
except ImportError:
    import configparser as ConfigParser

from sopel import bot, config, tools


__all__ = [
    'MockConfig',
    'MockSopel',
    'MockSopelWrapper',
    'get_example_test',
    'get_disable_setup',
    'insert_into_module',
    'run_example_tests',
]

if sys.version_info.major >= 3:
    basestring = str


class MockConfig(config.Config):
    @tools.deprecated('use configfactory fixture instead', '7.0', '8.0')
    def __init__(self):
        self.filename = tempfile.mkstemp()[1]
        self.parser = ConfigParser.RawConfigParser(allow_no_value=True)
        self.parser.add_section('core')
        self.parser.set('core', 'owner', 'Embolalia')
        self.define_section('core', config.core_section.CoreSection)
        self.get = self.parser.get

    def define_section(self, name, cls_):
        if not self.parser.has_section(name):
            self.parser.add_section(name)
        setattr(self, name, cls_(self, name))


class MockSopel(object):
    @tools.deprecated('use botfactory fixture instead', '7.0', '8.0')
    def __init__(self, nick, admin=False, owner=False):
        self.nick = nick
        self.user = "sopel"

        channel = tools.Identifier("#Sopel")
        self.channels = tools.SopelIdentifierMemory()
        self.channels[channel] = tools.target.Channel(channel)

        self.users = tools.SopelIdentifierMemory()
        self.privileges = tools.SopelMemory()

        self.memory = tools.SopelMemory()
        self.memory['url_callbacks'] = tools.SopelMemory()

        self.config = MockConfig()
        self._init_config()

        self.output = []

        if admin:
            self.config.core.admins = [self.nick]
        if owner:
            self.config.core.owner = self.nick

    def _store(self, string, *args, **kwargs):
        self.output.append(string.strip())

    write = msg = say = notice = action = reply = _store

    def _init_config(self):
        cfg = self.config
        cfg.parser.set('core', 'admins', '')
        cfg.parser.set('core', 'owner', '')
        home_dir = os.path.join(os.path.expanduser('~'), '.sopel')
        if not os.path.exists(home_dir):
            os.mkdir(home_dir)
        cfg.parser.set('core', 'homedir', home_dir)

    def register_url_callback(self, pattern, callback):
        if isinstance(pattern, basestring):
            pattern = re.compile(pattern)

        self.memory['url_callbacks'][pattern] = callback

    def unregister_url_callback(self, pattern, callback):
        if isinstance(pattern, basestring):
            pattern = re.compile(pattern)

        try:
            del self.memory['url_callbacks'][pattern]
        except KeyError:
            pass

    def search_url_callbacks(self, url):
        for regex, function in tools.iteritems(self.memory['url_callbacks']):
            match = regex.search(url)
            if match:
                yield function, match


class MockSopelWrapper(bot.SopelWrapper):
    @tools.deprecated('use sopel.bot.SopelWrapper instead', '7.0', '8.0')
    def __init__(self, *args, **kwargs):
        super(MockSopelWrapper, self).__init__(*args, **kwargs)


TEST_CONFIG = """
[core]
nick = {name}
owner = {owner}
admin = {admin}
"""


@tools.deprecated('this is now part of sopel.tests.pytest_plugin', '7.1', '8.0')
def get_example_test(*args, **kwargs):
    """Get a function that calls ``tested_func`` with fake wrapper and trigger.

    .. deprecated:: 7.1

        This is now part of the Sopel pytest plugin at
        :mod:`sopel.tests.pytest_plugin`.

    """
    from sopel.tests import pytest_plugin
    return pytest_plugin.get_example_test(*args, **kwargs)


@tools.deprecated('this is now part of sopel.tests.pytest_plugin', '7.1', '8.0')
def get_disable_setup():
    """Get a function to prevent conflict between pytest and plugin's setup.

    .. deprecated:: 7.1

        This is now part of the Sopel pytest plugin at
        :mod:`sopel.tests.pytest_plugin`.

    """
    from sopel.tests import pytest_plugin
    return pytest_plugin.get_disable_setup()


@tools.deprecated('this is now part of sopel.tests.pytest_plugin', '7.1', '8.0')
def insert_into_module(*args, **kwargs):
    """Add a function into a module.

    .. deprecated:: 7.1

        This is now part of the Sopel pytest plugin at
        :mod:`sopel.tests.pytest_plugin`.

    """
    from sopel.tests import pytest_plugin
    return pytest_plugin.insert_into_module(*args, **kwargs)


@tools.deprecated('pytest now runs @plugin.example tests directly', '7.1', '8.0')
def run_example_tests(filename, tb='native', multithread=False, verbose=False):
    # These are only required when running tests, so import them here rather
    # than at the module level.
    import pytest
    from multiprocessing import cpu_count

    args = [filename, "-s"]
    args.extend(['--tb', tb])
    if verbose:
        args.extend(['-v'])
    if multithread and cpu_count() > 1:
        args.extend(["-n", str(cpu_count())])

    pytest.main(args)



# coding=utf-8
"""
reddit.py - Sopel Reddit Plugin
Copyright 2012, Elsie Powell, embolalia.com
Copyright 2019, dgw, technobabbl.es
Copyright 2019, deathbybandaid, deathbybandaid.net
Licensed under the Eiffel Forum License 2.

https://sopel.chat
"""
from __future__ import absolute_import, division, print_function, unicode_literals

import datetime as dt
import re
import sys

import praw
import prawcore
import requests

from sopel import plugin
from sopel.formatting import bold, color, colors
from sopel.tools import time
from sopel.tools.web import USER_AGENT

# clean up all of this when dropping py2/old py3 versions
if sys.version_info.major >= 3:
    unicode = str
    if sys.version_info.minor >= 4:
        from html import unescape
    else:
        from html.parser import HTMLParser
        unescape = HTMLParser().unescape
else:
    from HTMLParser import HTMLParser
    unescape = HTMLParser().unescape

PLUGIN_OUTPUT_PREFIX = '[reddit] '

domain = r'https?://(?:www\.|old\.|pay\.|ssl\.|[a-z]{2}\.)?reddit\.com'
subreddit_url = r'%s/r/([\w-]+)/?$' % domain
post_or_comment_url = (
    domain +
    r'/r/\S+?/comments/(?P<submission>[\w-]+)'
    r'(?:/?(?:[\w%]+/(?P<comment>[\w-]+))?)'
)
short_post_url = r'https?://(redd\.it|reddit\.com)/(?P<submission>[\w-]+)'
user_url = r'%s/u(?:ser)?/([\w-]+)' % domain
image_url = r'https?://i\.redd\.it/\S+'
video_url = r'https?://v\.redd\.it/([\w-]+)'
gallery_url = r'https?://(?:www\.)?reddit\.com/gallery/([\w-]+)'


def setup(bot):
    if 'reddit_praw' not in bot.memory:
        # Create a PRAW instance just once, at load time
        bot.memory['reddit_praw'] = praw.Reddit(
            user_agent=USER_AGENT,
            client_id='6EiphT6SSQq7FQ',
            client_secret=None,
            check_for_updates=False,
        )


def shutdown(bot):
    # Clean up shared PRAW instance
    bot.memory.pop('reddit_praw', None)


def get_time_created(bot, trigger, entrytime):
    tz = time.get_timezone(
        bot.db, bot.config, None, trigger.nick, trigger.sender)
    time_created = dt.datetime.utcfromtimestamp(entrytime)
    created = time.format_time(bot.db,
                               bot.config, tz,
                               trigger.nick, trigger.sender,
                               time_created)
    return created


def get_is_cakeday(entrytime):
    now = dt.datetime.utcnow()
    cakeday_start = dt.datetime.utcfromtimestamp(entrytime)
    cakeday_start = cakeday_start.replace(year=now.year)
    day = dt.timedelta(days=1)
    year_div_by_400 = now.year % 400 == 0
    year_div_by_100 = now.year % 100 == 0
    year_div_by_4 = now.year % 4 == 0
    is_leap = year_div_by_400 or ((not year_div_by_100) and year_div_by_4)
    if (not is_leap) and ((cakeday_start.month, cakeday_start.day) == (2, 29)):
        # If cake day is 2/29 and it's not a leap year, cake day is 3/1.
        # Cake day begins at exact account creation time.
        is_cakeday = cakeday_start + day <= now <= cakeday_start + (2 * day)
    else:
        is_cakeday = cakeday_start <= now <= cakeday_start + day
    return is_cakeday


@plugin.url(image_url)
@plugin.output_prefix(PLUGIN_OUTPUT_PREFIX)
def image_info(bot, trigger, match):
    url = match.group(0)
    results = list(
        bot.memory['reddit_praw']
        .subreddit('all')
        .search('url:{}'.format(url), sort='new', params={'include_over_18': 'on'})
    )
    try:
        oldest = results[-1]
    except IndexError:
        # Fail silently if the image link can't be mapped to a submission
        return plugin.NOLIMIT
    return say_post_info(bot, trigger, oldest.id, False, True)


@plugin.url(video_url)
@plugin.output_prefix(PLUGIN_OUTPUT_PREFIX)
def video_info(bot, trigger, match):
    # Get the video URL with a cheeky hack
    url = requests.head(
        'https://www.reddit.com/video/{}'.format(match.group(1)),
        timeout=(10.0, 4.0)).headers['Location']
    try:
        return say_post_info(
            bot, trigger, re.match(post_or_comment_url, url).group('submission'), False, True
        )
    except AttributeError:
        # Fail silently if we can't map the video link to a submission
        return plugin.NOLIMIT


@plugin.url(post_or_comment_url)
@plugin.url(short_post_url)
@plugin.output_prefix(PLUGIN_OUTPUT_PREFIX)
def post_or_comment_info(bot, trigger, match):
    match = match or trigger
    groups = match.groupdict()

    if groups.get("comment"):
        say_comment_info(bot, trigger, groups["comment"])
        return

    say_post_info(bot, trigger, groups["submission"])


@plugin.url(gallery_url)
@plugin.output_prefix(PLUGIN_OUTPUT_PREFIX)
def rgallery_info(bot, trigger, match):
    match = match or trigger
    return say_post_info(bot, trigger, match.group(1), False)


def say_post_info(bot, trigger, id_, show_link=True, show_comments_link=False):
    try:
        s = bot.memory['reddit_praw'].submission(id=id_)

        message = (
            '{title} {link}{nsfw} | {points} {points_text} ({percent}) | '
            '{comments} {comments_text} | Posted by {author} | '
            'Created at {created}{comments_link}')

        subreddit = s.subreddit.display_name
        if not show_link:
            link = 'to r/{}'.format(subreddit)
        elif s.is_self:
            link = '(self.{})'.format(subreddit)
        else:
            link = '({}) to r/{}'.format(s.url, subreddit)

        nsfw = ''
        if s.over_18:
            nsfw += ' ' + bold(color('[NSFW]', colors.RED))

            sfw = bot.db.get_channel_value(trigger.sender, 'sfw')
            if sfw:
                link = '(link hidden)'
                bot.kick(
                    trigger.nick, trigger.sender,
                    'Linking to NSFW content in a SFW channel.'
                )
        if s.spoiler:
            nsfw += ' ' + bold(color('[SPOILER]', colors.GRAY))

            spoiler_free = bot.db.get_channel_value(trigger.sender, 'spoiler_free')
            if spoiler_free:
                link = '(link hidden)'
                bot.kick(
                    trigger.nick, trigger.sender,
                    'Linking to spoiler content in a spoiler-free channel.'
                )

        if s.author:
            author = s.author.name
        else:
            author = '[deleted]'

        created = get_time_created(bot, trigger, s.created_utc)

        if s.score > 0:
            point_color = colors.GREEN
        else:
            point_color = colors.RED

        points_text = 'point' if s.score == 1 else 'points'

        percent = color('{:.1%}'.format(s.upvote_ratio), point_color)

        comments_text = 'comment' if s.num_comments == 1 else 'comments'

        comments_link = ''
        if show_comments_link:
            try:
                comments_link = ' | ' + s.shortlink
            except AttributeError:
                # the value assigned earlier will be used
                pass

        title = unescape(s.title)
        message = message.format(
            title=title, link=link, nsfw=nsfw, points=s.score, points_text=points_text,
            percent=percent, comments=s.num_comments, comments_text=comments_text,
            author=author, created=created, comments_link=comments_link)

        bot.say(message)
    except prawcore.exceptions.NotFound:
        bot.reply('No such post.')
        return plugin.NOLIMIT


def say_comment_info(bot, trigger, id_):
    try:
        c = bot.memory['reddit_praw'].comment(id_)
    except prawcore.exceptions.NotFound:
        bot.reply('No such comment.')
        return plugin.NOLIMIT

    message = ('Comment by {author} | {points} {points_text} | '
               'Posted at {posted} | {comment}')

    if c.author:
        author = c.author.name
    else:
        author = '[deleted]'

    points_text = 'point' if c.score == 1 else 'points'

    posted = get_time_created(bot, trigger, c.created_utc)

    # stolen from the function I (dgw) wrote for our github plugin
    lines = [line for line in c.body.splitlines() if line and line[0] != '>']

    message = message.format(
        author=author, points=c.score, points_text=points_text,
        posted=posted, comment=' '.join(lines))

    bot.say(message, truncation=' […]')


def subreddit_info(bot, trigger, match, commanded=False):
    """Shows information about the given subreddit"""
    match_lower = match.lower()
    if match_lower in ['all', 'popular']:
        message = '{link}{nsfw} | {public_description}'
        nsfw = ' ' + bold(color('[Possible NSFW]', colors.ORANGE))
        link = "https://reddit.com/r/" + match_lower
        public_description = ''
        if match_lower == 'all':
            public_description = ('Today\'s top content from hundreds of '
                                  'thousands of Reddit communities.')
        elif match_lower == 'popular':
            public_description = ('The top trending content from some of '
                                  'Reddit\'s most popular communities')
        message = message.format(
            link=link, nsfw=nsfw, public_description=public_description)
        bot.say(message)
        return plugin.NOLIMIT

    r = bot.memory['reddit_praw']
    try:
        r.subreddits.search_by_name(match, exact=True)
    except prawcore.exceptions.NotFound:
        if commanded:
            bot.reply('No such subreddit.')
        # Fail silently if it wasn't an explicit command.
        return plugin.NOLIMIT

    try:
        s = r.subreddit(match)
        s.subreddit_type
    except prawcore.exceptions.Forbidden:
        bot.reply("r/" + match + " appears to be a private subreddit!")
        return plugin.NOLIMIT
    except prawcore.exceptions.NotFound:
        bot.reply("r/" + match + " appears to be a banned subreddit!")
        return plugin.NOLIMIT

    link = "https://reddit.com/r/" + s.display_name

    created = get_time_created(bot, trigger, s.created_utc)

    message = ('{link}{nsfw} | {subscribers} subscribers | '
               'Created at {created} | {public_description}')

    nsfw = ''
    if s.over18:
        nsfw += ' ' + bold(color('[NSFW]', colors.RED))

        sfw = bot.db.get_channel_value(trigger.sender, 'sfw')
        if sfw:
            link = '(link hidden)'
            bot.kick(
                trigger.nick, trigger.sender,
                'Linking to NSFW content in a SFW channel.'
            )

    message = message.format(
        link=link, nsfw=nsfw, subscribers='{:,}'.format(s.subscribers),
        created=created, public_description=s.public_description)

    bot.say(message, truncation=' […]')


def redditor_info(bot, trigger, match, commanded=False):
    """Shows information about the given Redditor"""
    try:
        u = bot.memory['reddit_praw'].redditor(match)
        u.id  # shortcut to check if the user exists or not
    except prawcore.exceptions.NotFound:
        if commanded:
            bot.reply('No such Redditor.')
        # Fail silently if it wasn't an explicit command.
        return plugin.NOLIMIT

    message = u.name
    is_cakeday = get_is_cakeday(u.created_utc)

    if is_cakeday:
        message = message + ' | ' + bold(color('Cake day', colors.LIGHT_PURPLE))
    if commanded:
        message = message + ' | https://reddit.com/u/' + u.name
    if u.is_gold:
        message = message + ' | ' + bold(color('Gold', colors.YELLOW))
    if u.is_employee:
        message = message + ' | ' + bold(color('Employee', colors.RED))
    if u.is_mod:
        message = message + ' | ' + bold(color('Mod', colors.GREEN))
    message = message + (' | Link: ' + str(u.link_karma) +
                         ' | Comment: ' + str(u.comment_karma))

    bot.say(message)


@plugin.url(user_url)
@plugin.output_prefix(PLUGIN_OUTPUT_PREFIX)
def auto_redditor_info(bot, trigger, match):
    return redditor_info(bot, trigger, match.group(1))


@plugin.url(subreddit_url)
@plugin.output_prefix(PLUGIN_OUTPUT_PREFIX)
def auto_subreddit_info(bot, trigger, match):
    return subreddit_info(bot, trigger, match.group(1))


@plugin.require_chanmsg('Setting SFW status is only supported in a channel.')
@plugin.require_privilege(plugin.OP)
@plugin.command('setsafeforwork', 'setsfw')
@plugin.example('.setsfw true')
@plugin.example('.setsfw false')
@plugin.output_prefix(PLUGIN_OUTPUT_PREFIX)
def set_channel_sfw(bot, trigger):
    """
    Sets the Safe for Work status (true or false) for the current
    channel. Defaults to false.
    """
    param = 'true'
    if trigger.group(2) and trigger.group(3):
        param = trigger.group(3).strip().lower()
    sfw = param == 'true'
    bot.db.set_channel_value(trigger.sender, 'sfw', sfw)
    if sfw:
        bot.say('%s is now flagged as SFW.' % trigger.sender)
    else:
        bot.say('%s is now flagged as NSFW.' % trigger.sender)


@plugin.command('getsafeforwork', 'getsfw')
@plugin.example('.getsfw [channel]')
@plugin.output_prefix(PLUGIN_OUTPUT_PREFIX)
def get_channel_sfw(bot, trigger):
    """
    Gets the preferred channel's Safe for Work status, or the current
    channel's status if no channel given.
    """
    channel = trigger.group(2)
    if not channel:
        channel = trigger.sender
        if channel.is_nick():
            bot.reply('{}getsfw with no channel param is only permitted in '
                      'channels.'.format(bot.config.core.help_prefix))
            return

    channel = channel.strip()

    sfw = bot.db.get_channel_value(channel, 'sfw')
    if sfw:
        bot.say('%s is flagged as SFW' % channel)
    else:
        bot.say('%s is flagged as NSFW' % channel)


@plugin.require_chanmsg('Only channels can be marked as spoiler-free.')
@plugin.require_privilege(plugin.OP)
@plugin.command('setspoilerfree', 'setspoilfree')
@plugin.example('.setspoilfree true')
@plugin.example('.setspoilfree false')
@plugin.output_prefix(PLUGIN_OUTPUT_PREFIX)
def set_channel_spoiler_free(bot, trigger):
    """
    Sets the Spoiler-Free status (true or false) for the current channel.
    Defaults to false.
    """
    param = 'true'
    if trigger.group(2) and trigger.group(3):
        param = trigger.group(3).strip().lower()
    spoiler_free = param == 'true'
    bot.db.set_channel_value(trigger.sender, 'spoiler_free', spoiler_free)
    if spoiler_free:
        bot.say('%s is now flagged as spoiler-free.' % trigger.sender)
    else:
        bot.say('%s is now flagged as spoilers-allowed.' % trigger.sender)


@plugin.command('getspoilerfree', 'getspoilfree')
@plugin.example('.getspoilfree [channel]')
@plugin.output_prefix(PLUGIN_OUTPUT_PREFIX)
def get_channel_spoiler_free(bot, trigger):
    """
    Gets the channel's Spoiler-Free status, or the current channel's
    status if no channel given.
    """
    channel = trigger.group(2)
    if not channel:
        channel = trigger.sender
        if channel.is_nick():
            bot.reply('{}getspoilfree with no channel param is only permitted '
                      'in channels.'.format(bot.config.core.help_prefix))
            return

    channel = channel.strip()

    spoiler_free = bot.db.get_channel_value(channel, 'spoiler_free')
    if spoiler_free:
        bot.say('%s is flagged as spoiler-free' % channel)
    else:
        bot.say('%s is flagged as spoilers-allowed' % channel)


@plugin.find(r'(?<!\S)/?(?P<prefix>r|u)/(?P<id>[a-zA-Z0-9-_]+)\b')
@plugin.output_prefix(PLUGIN_OUTPUT_PREFIX)
def reddit_slash_info(bot, trigger):
    searchtype = trigger.group('prefix').lower()
    match = trigger.group('id')
    if searchtype == "r":
        return subreddit_info(bot, trigger, match, commanded=False)
    elif searchtype == "u":
        return redditor_info(bot, trigger, match, commanded=False)


@plugin.command('subreddit')
@plugin.example('.subreddit plex')
@plugin.output_prefix(PLUGIN_OUTPUT_PREFIX)
def subreddit_command(bot, trigger):
    # require input
    if not trigger.group(2):
        bot.reply('You must provide a subreddit name.')
        return

    # subreddit names do not contain spaces
    match = trigger.group(3)
    return subreddit_info(bot, trigger, match, commanded=True)


@plugin.command('redditor')
@plugin.example('.redditor poem_for_your_sprog')
@plugin.output_prefix(PLUGIN_OUTPUT_PREFIX)
def redditor_command(bot, trigger):
    # require input
    if not trigger.group(2):
        bot.reply('You must provide a Redditor name.')
        return

    # Redditor names do not contain spaces
    match = trigger.group(3)
    return redditor_info(bot, trigger, match, commanded=True)



# coding=utf-8
"""IRC Tools for ISUPPORT management.

When a server wants to advertise its features and settings, it can use the
``RPL_ISUPPORT`` command (``005`` numeric) with a list of arguments.

.. seealso::

    https://modern.ircdocs.horse/#rplisupport-005

"""
# Copyright 2019, Florian Strzelecki <florian.strzelecki@gmail.com>
#
# Licensed under the Eiffel Forum License 2.
from __future__ import absolute_import, division, print_function, unicode_literals

import functools
import itertools
import re


def _optional(parser, default=None):
    # set a parser as optional: will always return the default value provided
    # if there is no value (empty or None)
    @functools.wraps(parser)
    def wrapped(value):
        if not value:
            return default
        return parser(value)
    return wrapped


def _no_value(value):
    # always ignore the value
    return None


def _single_character(value):
    if len(value) > 1:
        raise ValueError('Too many characters: %r.' % value)

    return value


def _map_items(parser=str, map_separator=',', item_separator=':'):
    @functools.wraps(parser)
    def wrapped(value):
        items = sorted(
            item.split(item_separator)
            for item in value.split(map_separator))

        return tuple(
            (k, parser(v) if v else None)
            for k, v in items
        )
    return wrapped


def _parse_chanmodes(value):
    items = value.split(',')

    if len(items) < 4:
        raise ValueError('Not enough channel types to unpack from %r.' % value)

    # add extra channel mode types to their own tuple
    # result in (A, B, C, D, (E, F, G, H, ..., Z))
    # where A, B, C, D = result[:4]
    # and extras = result[4]
    return tuple(items[:4]) + (tuple(items[4:]),)


def _parse_elist(value):
    # letters are case-insensitives
    return tuple(sorted(set(letter.upper() for letter in value)))


def _parse_extban(value):
    args = value.split(',')

    if len(args) < 2:
        raise ValueError('Invalid value for EXTBAN: %r.' % value)

    prefix = args[0] or None
    items = tuple(sorted(set(args[1])))

    return (prefix, items)


def _parse_prefix(value):
    result = re.match(r'\((?P<modes>\S+)\)(?P<prefixes>\S+)', value)

    if not result:
        raise ValueError('Invalid value for PREFIX: %r' % value)

    modes = result.group('modes')
    prefixes = result.group('prefixes')

    if len(modes) != len(prefixes):
        raise ValueError('Mode list does not match for PREFIX: %r' % value)

    return tuple(sorted(zip(modes, prefixes)))


ISUPPORT_PARSERS = {
    'AWAYLEN': int,
    'CASEMAPPING': str,
    'CHANLIMIT': _map_items(int),
    'CHANMODES': _parse_chanmodes,
    'CHANNELLEN': int,
    'CHANTYPES': _optional(tuple),
    'ELIST': _parse_elist,
    'EXCEPTS': _optional(_single_character, default='e'),
    'EXTBAN': _parse_extban,
    'HOSTLEN': int,
    'INVEX': _optional(_single_character, default='I'),
    'KICKLEN': int,
    'MAXLIST': _map_items(int),
    'MAXTARGETS': _optional(int),
    'MODES': _optional(int),
    'NETWORK': str,
    'NICKLEN': int,
    'PREFIX': _optional(_parse_prefix),
    'SAFELIST': _no_value,
    'SILENCE': _optional(int),
    'STATUSMSG': _optional(tuple),
    'TARGMAX': _optional(_map_items(int), default=tuple()),
    'TOPICLEN': int,
    'USERLEN': int,
}


def parse_parameter(arg):
    items = arg.split('=', 1)
    if len(items) == 2:
        key, value = items
    else:
        key, value = items[0], None

    if key.startswith('-'):
        # ignore value for removed parameters
        return (key, None)

    parser = ISUPPORT_PARSERS.get(key, _optional(str))
    return (key, parser(value))


class ISupport(object):
    """Storage class for IRC's ``ISUPPORT`` feature.

    An instance of ``ISupport`` can be used as a read-only dict, to store
    features advertised by the IRC server::

        >>> isupport = ISupport(chanlimit=(('&', None), ('#', 70)))
        >>> isupport['CHANLIMIT']
        (('&', None) ('#', 70))
        >>> isupport.CHANLIMIT  # some parameters are also properties
        {
            '&': None,
            '#': 70,
        }
        >>> 'chanlimit' in isupport  # case-insensitive
        True
        >>> 'chanmode' in isupport
        False
        >>> isupport.CHANMODE  # not advertised by the server!
        Traceback (most recent call last):
          File "<stdin>", line 1, in <module>
        AttributeError: 'ISupport' object has no attribute 'CHANMODE'

    The list of possible parameters can be found at
    `modern.ircdocs.horse's RPL_ISUPPORT Parameters`__.

    .. __: https://modern.ircdocs.horse/#rplisupport-parameters
    """
    def __init__(self, **kwargs):
        self.__isupport = dict(
            (key.upper(), value)
            for key, value in kwargs.items()
            if not key.startswith('-'))

    def __getitem__(self, key):
        key_ci = key.upper()
        if key_ci not in self.__isupport:
            raise KeyError(key_ci)
        return self.__isupport[key_ci]

    def __contains__(self, key):
        return key.upper() in self.__isupport

    def __getattr__(self, name):
        if name not in self.__isupport:
            raise AttributeError(name)

        return self.__isupport[name]

    def __setattr__(self, name, value):
        # make sure you can't set the value of any ISUPPORT attribute yourself
        if name == '_ISupport__isupport':
            # allow to set self.__isupport inside of the class
            super(ISupport, self).__setattr__(name, value)
        elif name in self.__isupport:
            # reject any modification of __isupport
            raise AttributeError("Can't set value for %r" % name)
        elif name not in self.__dict__:
            raise AttributeError('Unknown attribute')

    def get(self, name, default=None):
        """Retrieve value for the feature ``name``.

        :param str name: feature to retrieve
        :param default: default value if the feature is not advertised
                        (defaults to ``None``)
        :return: the value for that feature, if advertised, or ``default``
        """
        return self[name] if name in self else default

    def apply(self, **kwargs):
        """Build a new instance of :class:`ISupport`.

        :return: a new instance, updated with the latest advertised features
        :rtype: :class:`ISupport`

        This method applies the latest advertised features from the server:
        the result contains the new and updated parameters, and doesn't contain
        the removed parameters (marked by ``-{PARAMNAME}``)::

            >>> updated = {'-AWAYLEN': None, 'NICKLEN': 25, 'CHANNELLEN': 10}
            >>> new = isupport.apply(**updated)
            >>> 'CHANNELLEN' in new
            True
            >>> 'AWAYLEN' in new
            False

        """
        kwargs_upper = dict(
            (key.upper(), value)
            for key, value in kwargs.items()
        )
        kept = (
            (key, value)
            for key, value in self.__isupport.items()
            if ('-%s' % key) not in kwargs_upper
        )
        updated = dict(itertools.chain(kept, kwargs_upper.items()))

        return self.__class__(**updated)

    @property
    def CHANLIMIT(self):
        """Expose ``CHANLIMIT`` as a dict, if advertised by the server.

        This exposes information about the maximum number of channels that the
        bot can join for each prefix::

            >>> isupport.CHANLIMIT
            {
                '#': 70,
                '&': None,
            }

        In that example, the bot may join 70 ``#`` channels and any number of
        ``&`` channels.

        This attribute is not available if the server does not provide the
        right information, and accessing it will raise an
        :exc:`AttributeError`.

        .. seealso::

            https://modern.ircdocs.horse/#chanlimit-parameter

        """
        if 'CHANLIMIT' not in self:
            raise AttributeError('CHANLIMIT')

        return dict(self['CHANLIMIT'])

    @property
    def CHANMODES(self):
        """Expose ``CHANMODES`` as a dict.

        This exposes information about 4 types of channel modes::

            >>> isupport.CHANMODES
            {
                'A': 'b',
                'B': 'k',
                'C': 'l',
                'D': 'imnpst',
            }

        The values are empty if the server does not provide this information.

        .. seealso::

            https://modern.ircdocs.horse/#chanmodes-parameter

        """
        if 'CHANMODES' not in self:
            return {"A": "", "B": "", "C": "", "D": ""}

        return dict(zip('ABCD', self['CHANMODES'][:4]))

    @property
    def MAXLIST(self):
        """Expose ``MAXLIST`` as a dict, if advertised by the server.

        This exposes information about maximums for combinations of modes::

            >>> isupport.MAXLIST
            {
                'beI': 100,
                'q': 50,
                'b': 50,
            }

        This attribute is not available if the server does not provide the
        right information, and accessing it will raise an
        :exc:`AttributeError`.

        .. seealso::

            https://modern.ircdocs.horse/#maxlist-parameter

        """
        if 'MAXLIST' not in self:
            raise AttributeError('MAXLIST')

        return dict(self['MAXLIST'])

    @property
    def PREFIX(self):
        """Expose ``PREFIX`` as a dict, if advertised by the server.

        This exposes information about the modes and nick prefixes used for
        user privileges in channels::

            >>> isupport.PREFIX
            {
                'q': '~',
                'a': '&',
                'o': '@',
                'h': '%',
                'v': '+',
            }

        This attribute is not available if the server does not provide the
        right information, and accessing it will raise an
        :exc:`AttributeError`.

        .. seealso::

            https://modern.ircdocs.horse/#prefix-parameter

        """
        if 'PREFIX' not in self:
            raise AttributeError('PREFIX')

        return dict(self['PREFIX'])

    @property
    def TARGMAX(self):
        """Expose ``TARGMAX`` as a dict, if advertised by the server.

        This exposes information about the maximum number of arguments for
        each command::

            >>> isupport.TARGMAX
            {
                'JOIN': None,
                'PRIVMSG': 3,
                'WHOIS': 1,
            }
            >>> isupport['TARGMAX']  # internal representation
            (('JOIN', None), ('PRIVMSG', 3), ('WHOIS', 1))

        This attribute is not available if the server does not provide the
        right information, and accessing it will raise an
        :exc:`AttributeError`.

        The internal representation of ``TARGMAX`` is a tuple of 2-value
        tuples as seen above.

        .. seealso::

            https://modern.ircdocs.horse/#targmax-parameter

        """
        if 'TARGMAX' not in self:
            raise AttributeError('TARGMAX')

        # always return a dict if None or empty tuple
        return dict(self['TARGMAX'] or [])



# coding=utf-8
"""Tests for core ``sopel.irc``"""
from __future__ import absolute_import, division, print_function, unicode_literals

import pytest

from sopel.tests import rawlist


TMP_CONFIG = """
[core]
owner = Exirel
nick = Sopel
user = sopel
name = Sopel (https://sopel.chat)
# we don't want flood protection here
flood_burst_lines = 1000
"""


@pytest.fixture
def tmpconfig(configfactory):
    return configfactory('conf.ini', TMP_CONFIG)


@pytest.fixture
def bot(tmpconfig, botfactory):
    return botfactory(tmpconfig)


def prefix_length(bot):
    # ':', nick, '!', '~', ident/username, '@', maximum hostname length, <0x20>
    return 1 + len(bot.nick) + 1 + 1 + len(bot.user) + 1 + 63 + 1


def test_on_connect(bot):
    bot.on_connect()

    assert bot.backend.message_sent == rawlist(
        'CAP LS 302',
        'NICK Sopel',
        'USER sopel 0 * :Sopel (https://sopel.chat)'
    )


def test_on_connect_auth_password(bot):
    bot.settings.core.auth_method = 'server'
    bot.settings.core.auth_password = 'auth_secret'
    bot.on_connect()

    assert bot.backend.message_sent == rawlist(
        'CAP LS 302',
        'PASS auth_secret',
        'NICK Sopel',
        'USER sopel 0 * :Sopel (https://sopel.chat)'
    )


def test_on_connect_server_auth_password(bot):
    bot.settings.core.server_auth_method = 'server'
    bot.settings.core.server_auth_password = 'server_secret'
    bot.on_connect()

    assert bot.backend.message_sent == rawlist(
        'CAP LS 302',
        'PASS server_secret',
        'NICK Sopel',
        'USER sopel 0 * :Sopel (https://sopel.chat)'
    )


def test_on_connect_auth_password_override_server_auth(bot):
    bot.settings.core.auth_method = 'server'
    bot.settings.core.auth_password = 'auth_secret'
    bot.settings.core.server_auth_method = 'server'
    bot.settings.core.server_auth_password = 'server_secret'
    bot.on_connect()

    assert bot.backend.message_sent == rawlist(
        'CAP LS 302',
        'PASS auth_secret',
        'NICK Sopel',
        'USER sopel 0 * :Sopel (https://sopel.chat)'
    )


def test_write(bot):
    bot.write(['INFO'])

    assert bot.backend.message_sent == rawlist('INFO')


def test_write_args(bot):
    bot.write(['NICK', 'Sopel'])

    assert bot.backend.message_sent == rawlist('NICK Sopel')


def test_write_text(bot):
    bot.write(['HELP'], '?')

    assert bot.backend.message_sent == rawlist('HELP :?')


def test_write_args_text_safe(bot):
    bot.write(['CMD\nUNSAFE'], 'Unsafe\rtext')

    assert bot.backend.message_sent == rawlist('CMDUNSAFE :Unsafetext')


def test_write_args_many(bot):
    bot.write(['NICK', 'Sopel'])
    bot.write(['JOIN', '#sopel'])

    assert bot.backend.message_sent == rawlist(
        'NICK Sopel',
        'JOIN #sopel',
    )


def test_write_text_many(bot):
    bot.write(['NICK', 'Sopel'])
    bot.write(['HELP'], '?')

    assert bot.backend.message_sent == rawlist(
        'NICK Sopel',
        'HELP :?',
    )


def test_action(bot):
    bot.action('is doing some tests', '#sopel')

    assert bot.backend.message_sent == rawlist(
        'PRIVMSG #sopel :\001ACTION is doing some tests\001',
    )


def test_join(bot):
    bot.join('#sopel')

    assert bot.backend.message_sent == rawlist(
        'JOIN #sopel',
    )


def test_join_password(bot):
    bot.join('#sopel', 'secret_password')

    assert bot.backend.message_sent == rawlist(
        'JOIN #sopel secret_password',
    )


def test_kick(bot):
    bot.kick('spambot', '#channel')

    assert bot.backend.message_sent == rawlist(
        'KICK #channel spambot',
    )


def test_kick_reason(bot):
    bot.kick('spambot', '#channel', 'Flood!')

    assert bot.backend.message_sent == rawlist(
        'KICK #channel spambot :Flood!',
    )


def test_notice(bot):
    bot.notice('Hello world!', '#sopel')

    assert bot.backend.message_sent == rawlist(
        'NOTICE #sopel :Hello world!',
    )


def test_part(bot):
    bot.part('#channel')

    assert bot.backend.message_sent == rawlist(
        'PART #channel',
    )


def test_part_reason(bot):
    bot.part('#channel', 'Bye!')

    assert bot.backend.message_sent == rawlist(
        'PART #channel :Bye!',
    )


def test_reply(bot):
    bot.reply('Thank you!', '#sopel', 'dgw')

    assert bot.backend.message_sent == rawlist(
        'PRIVMSG #sopel :dgw: Thank you!',
    )


def test_reply_notice(bot):
    bot.reply('Thank you!', '#sopel', 'dgw', notice=True)

    assert bot.backend.message_sent == rawlist(
        'NOTICE #sopel :dgw: Thank you!',
    )


def test_say(bot):
    bot.say('Hello world!', '#sopel')

    assert bot.backend.message_sent == rawlist(
        'PRIVMSG #sopel :Hello world!',
    )


def test_say_safe(bot):
    bot.say('Hello\r\nworld!\r\n', '#sopel')

    assert bot.backend.message_sent == rawlist(
        'PRIVMSG #sopel :Helloworld!',
    )


def test_say_long_fit(bot):
    """Test a long message that fits into the 512 bytes limit."""
    text = 'a' * (512 - len('PRIVMSG #sopel :\r\n'))
    bot.say(text, '#sopel')

    assert bot.backend.message_sent == rawlist(
        'PRIVMSG #sopel :%s' % text,
    )


def test_say_long_extra(bot):
    """Test a long message that doesn't fit into the 512 bytes limit."""
    text = 'a' * (512 - len('PRIVMSG #sopel :\r\n'))
    bot.say(text + 'b', '#sopel')

    assert bot.backend.message_sent == rawlist(
        'PRIVMSG #sopel :%s' % text,  # the 'b' is truncated out
    )


def test_say_long_extra_multi_message(bot):
    """Test a long message that doesn't fit, with split allowed."""
    text = 'a' * (512 - prefix_length(bot) - len('PRIVMSG #sopel :\r\n'))
    bot.say(text + 'b', '#sopel', max_messages=2)

    assert bot.backend.message_sent == rawlist(
        'PRIVMSG #sopel :%s' % text,  # the 'b' is split from message
        'PRIVMSG #sopel :b',
    )


def test_say_long_extra_multi_message_multibyte_recipient(bot):
    """Test a split-allowed message sent to recipient with multi-byte char."""
    text = 'a' * (
        512 - prefix_length(bot) - len('PRIVMSG #sopel² :\r\n'.encode('utf-8')))
    bot.say(text + 'b', '#sopel²', max_messages=2)

    assert bot.backend.message_sent == rawlist(
        'PRIVMSG #sopel² :%s' % text,  # the 'b' is split from message
        'PRIVMSG #sopel² :b',
    )


def test_say_long_truncation_fit(bot):
    """Test optional truncation indicator with message that fits in one line."""
    text = 'a' * (512 - prefix_length(bot) - len('PRIVMSG #sopel :\r\n') - 3)
    bot.say(text + 'ttt', '#sopel', truncation='...')

    assert bot.backend.message_sent == rawlist(
        'PRIVMSG #sopel :%s' % text + 'ttt',  # nothing is truncated or replaced
    )


def test_say_long_truncation_extra(bot):
    """Test optional truncation indicator with message that is too long."""
    text = 'a' * (512 - prefix_length(bot) - len('PRIVMSG #sopel :\r\n') - 3)
    bot.say(text + 'ttt' + 'b', '#sopel', truncation='...')

    assert bot.backend.message_sent == rawlist(
        # 'b' is truncated; 'ttt' is replaced by `truncation`
        'PRIVMSG #sopel :%s' % text + '...',
    )


def test_say_long_truncation_extra_multi_message(bot):
    """Test optional truncation indicator with message splitting."""
    msg1 = 'a' * (512 - prefix_length(bot) - len('PRIVMSG #sopel :\r\n'))
    msg2 = msg1[:-3] + 'ttt'
    bot.say(msg1 + msg2 + 'b', '#sopel', max_messages=2, truncation='...')

    assert bot.backend.message_sent == rawlist(
        # split as expected
        'PRIVMSG #sopel :%s' % msg1,
        # 'b' is truncated; 'ttt' is replaced by `truncation`
        'PRIVMSG #sopel :%s' % msg2.replace('ttt', '...'),
    )


def test_say_long_truncation_extra_multi_message_multibyte(bot):
    """Test optional truncation indicator with message splitting."""
    msg1 = 'a' * (512 - prefix_length(bot) - len('PRIVMSG #sopel :\r\n'))
    msg2 = msg1[:-3] + 'ttt'
    bot.say(msg1 + msg2 + 'b', '#sopel', max_messages=2, truncation='…')

    assert bot.backend.message_sent == rawlist(
        # split as expected
        'PRIVMSG #sopel :%s' % msg1,
        # 'b' is truncated; 'ttt' is replaced by `truncation`
        'PRIVMSG #sopel :%s' % msg2.replace('ttt', '…'),
    )


def test_say_trailing(bot):
    """Test optional trailing string."""
    text = '"This is a test quote.'
    bot.say(text, '#sopel', trailing='"')

    assert bot.backend.message_sent == rawlist(
        # combined
        'PRIVMSG #sopel :%s' % text + '"'
    )


def test_say_long_truncation_trailing(bot):
    """Test optional truncation indicator AND trailing string together."""
    msg1 = 'a' * (512 - prefix_length(bot) - len('PRIVMSG #sopel :\r\n'))
    msg2 = msg1[:-4] + 'bbbc'
    bot.say(msg1 + msg2 + 'd', '#sopel', max_messages=2, truncation='…', trailing='q')

    assert bot.backend.message_sent == rawlist(
        # split as expected
        'PRIVMSG #sopel :%s' % msg1,
        # 'd' is truncated; 'bbb' is replaced by `truncation`; 'c' is replaced by `trailing`
        'PRIVMSG #sopel :%s' % msg2.replace('bbb', '…').replace('c', 'q')
    )


def test_say_no_repeat_protection(bot):
    # five is fine
    bot.say('hello', '#sopel')
    bot.say('hello', '#sopel')
    bot.say('hello', '#sopel')
    bot.say('hello', '#sopel')
    bot.say('hello', '#sopel')

    assert bot.backend.message_sent == rawlist(
        'PRIVMSG #sopel :hello',
        'PRIVMSG #sopel :hello',
        'PRIVMSG #sopel :hello',
        'PRIVMSG #sopel :hello',
        'PRIVMSG #sopel :hello',
    )

    # six: replaced by '...'
    bot.say('hello', '#sopel')

    assert bot.backend.message_sent == rawlist(
        'PRIVMSG #sopel :hello',
        'PRIVMSG #sopel :hello',
        'PRIVMSG #sopel :hello',
        'PRIVMSG #sopel :hello',
        'PRIVMSG #sopel :hello',
        # the extra hello is replaced by '...'
        'PRIVMSG #sopel :...',
    )

    # these one will add more '...'
    bot.say('hello', '#sopel')
    bot.say('hello', '#sopel')

    assert bot.backend.message_sent == rawlist(
        'PRIVMSG #sopel :hello',
        'PRIVMSG #sopel :hello',
        'PRIVMSG #sopel :hello',
        'PRIVMSG #sopel :hello',
        'PRIVMSG #sopel :hello',
        'PRIVMSG #sopel :...',
        # the new ones are also replaced by '...'
        'PRIVMSG #sopel :...',
        'PRIVMSG #sopel :...',
    )

    # but at some point it just stops talking
    bot.say('hello', '#sopel')

    assert bot.backend.message_sent == rawlist(
        'PRIVMSG #sopel :hello',
        'PRIVMSG #sopel :hello',
        'PRIVMSG #sopel :hello',
        'PRIVMSG #sopel :hello',
        'PRIVMSG #sopel :hello',
        #  three time, then stop
        'PRIVMSG #sopel :...',
        'PRIVMSG #sopel :...',
        'PRIVMSG #sopel :...',
    )



# coding=utf-8
"""
version.py - Sopel Version Plugin
Copyright 2009, Silas Baronda
Copyright 2014, Dimitri Molenaars <tyrope@tyrope.nl>
Licensed under the Eiffel Forum License 2.

https://sopel.chat
"""
from __future__ import absolute_import, division, print_function, unicode_literals

import datetime
import os
import platform

from sopel import __version__ as release, plugin


PROJECT_DIR = os.path.dirname(os.path.dirname(os.path.dirname(__file__)))
GIT_DIR = os.path.join(PROJECT_DIR, '.git')


def git_info():
    head = os.path.join(GIT_DIR, 'HEAD')
    if os.path.isfile(head):
        with open(head) as h:
            head_loc = h.readline()[5:-1]  # strip ref: and \n
        head_file = os.path.join(GIT_DIR, head_loc)
        if os.path.isfile(head_file):
            with open(head_file) as h:
                sha = h.readline()
                if sha:
                    return sha


@plugin.command('version')
@plugin.output_prefix('[version] ')
def version(bot, trigger):
    """Display the installed version of Sopel.

    Includes the version of Python Sopel is installed on.
    Includes the commit hash if Sopel is installed from source.
    """
    parts = [
        'Sopel v%s' % release,
        'Python: %s' % platform.python_version()
    ]
    sha = git_info()
    if sha:
        parts.append('Commit: %s' % sha)

    bot.say(' | '.join(parts))


@plugin.ctcp('VERSION')
@plugin.rate(20)
def ctcp_version(bot, trigger):
    bot.write(('NOTICE', trigger.nick),
              '\x01VERSION Sopel IRC Bot version %s\x01' % release)


@plugin.ctcp('SOURCE')
@plugin.rate(20)
def ctcp_source(bot, trigger):
    bot.write(('NOTICE', trigger.nick),
              '\x01SOURCE https://github.com/sopel-irc/sopel\x01')


@plugin.ctcp('PING')
@plugin.rate(10)
def ctcp_ping(bot, trigger):
    text = trigger.group()
    text = text.replace("PING ", "")
    text = text.replace("\x01", "")
    bot.write(('NOTICE', trigger.nick),
              '\x01PING {0}\x01'.format(text))


@plugin.ctcp('TIME')
@plugin.rate(20)
def ctcp_time(bot, trigger):
    dt = datetime.datetime.now()
    current_time = dt.strftime("%A, %d. %B %Y %I:%M%p")
    bot.write(('NOTICE', trigger.nick),
              '\x01TIME {0}\x01'.format(current_time))



# coding=utf-8
"""Tests for the new database functionality.

TODO: Most of these tests assume functionality tested in other tests. This is
enough to get everything working (and is better than nothing), but best
practice would probably be not to do that."""
from __future__ import absolute_import, division, print_function, unicode_literals

import json
import os
import sys
import tempfile

import pytest

from sopel.db import ChannelValues, Nicknames, NickValues, PluginValues, SopelDB
from sopel.tools import Identifier

db_filename = tempfile.mkstemp()[1]
if sys.version_info.major >= 3:
    unicode = str
    basestring = str
    iteritems = dict.items
    itervalues = dict.values
    iterkeys = dict.keys
else:
    iteritems = dict.iteritems
    itervalues = dict.itervalues
    iterkeys = dict.iterkeys


TMP_CONFIG = """
[core]
owner = Embolalia
db_filename = {db_filename}
"""


@pytest.fixture
def db(configfactory):
    content = TMP_CONFIG.format(db_filename=db_filename)
    settings = configfactory('default.cfg', content)
    db = SopelDB(settings)
    # TODO add tests to ensure db creation works properly, too.
    return db


def teardown_function(function):
    os.remove(db_filename)


def test_get_nick_id(db):
    session = db.ssession()
    tests = [
        [None, 'embolalia', Identifier('Embolalia')],
        # Ensures case conversion is handled properly
        [None, '{}{}', Identifier('[]{}')],
        # Unicode, just in case
        [None, 'embölaliå', Identifier('EmbölaliÅ')],
    ]

    for test in tests:
        test[0] = db.get_nick_id(test[2])
        nick_id, slug, nick = test
        registered = session.query(Nicknames) \
                            .filter(Nicknames.canonical == nick) \
                            .all()
        assert len(registered) == 1
        assert registered[0].slug == slug and registered[0].canonical == nick

    # Check that each nick ended up with a different id
    assert len(set(test[0] for test in tests)) == len(tests)

    # Check that the retrieval actually is idempotent
    for test in tests:
        nick_id = test[0]
        new_id = db.get_nick_id(test[2])
        assert nick_id == new_id

    # Even if the case is different
    for test in tests:
        nick_id = test[0]
        new_id = db.get_nick_id(Identifier(test[2].upper()))
        assert nick_id == new_id
    session.close()


def test_alias_nick(db):
    nick = 'Embolalia'
    aliases = ['EmbölaliÅ', 'Embo`work', 'Embo']

    nick_id = db.get_nick_id(nick)
    for alias in aliases:
        db.alias_nick(nick, alias)

    for alias in aliases:
        assert db.get_nick_id(alias) == nick_id

    db.alias_nick('both', 'arenew')  # Shouldn't fail.

    with pytest.raises(ValueError):
        db.alias_nick('Eve', nick)

    with pytest.raises(ValueError):
        db.alias_nick(nick, nick)


def test_set_nick_value(db):
    session = db.ssession()
    nick = 'Embolalia'
    nick_id = db.get_nick_id(nick)
    data = {
        'key': 'value',
        'number_key': 1234,
        'unicode': 'EmbölaliÅ',
    }

    def check():
        for key, value in iteritems(data):
            db.set_nick_value(nick, key, value)

        for key, value in iteritems(data):
            found_value = session.query(NickValues.value) \
                                 .filter(NickValues.nick_id == nick_id) \
                                 .filter(NickValues.key == key) \
                                 .scalar()
            assert json.loads(unicode(found_value)) == value
    check()

    # Test updates
    data['number_key'] = 'not a number anymore!'
    data['unicode'] = 'This is different toö!'
    check()
    session.close()


def test_get_nick_value(db):
    session = db.ssession()
    nick = 'Embolalia'
    nick_id = db.get_nick_id(nick)
    data = {
        'key': 'value',
        'number_key': 1234,
        'unicode': 'EmbölaliÅ',
    }

    for key, value in iteritems(data):
        nv = NickValues(nick_id=nick_id, key=key, value=json.dumps(value, ensure_ascii=False))
        session.add(nv)
        session.commit()

    for key, value in iteritems(data):
        found_value = db.get_nick_value(nick, key)
        assert found_value == value
    session.close()


def test_get_nick_value_default(db):
    assert db.get_nick_value("TestUser", "DoesntExist") is None
    assert db.get_nick_value("TestUser", "DoesntExist", "MyDefault") == "MyDefault"


def test_delete_nick_value(db):
    nick = 'Embolalia'
    db.set_nick_value(nick, 'wasd', 'uldr')
    assert db.get_nick_value(nick, 'wasd') == 'uldr'
    db.delete_nick_value(nick, 'wasd')
    assert db.get_nick_value(nick, 'wasd') is None


def test_unalias_nick(db):
    session = db.ssession()
    nick = 'Embolalia'
    nick_id = 42

    nn = Nicknames(nick_id=nick_id, slug=Identifier(nick).lower(), canonical=nick)
    session.add(nn)
    session.commit()

    aliases = ['EmbölaliÅ', 'Embo`work', 'Embo']
    for alias in aliases:
        nn = Nicknames(nick_id=nick_id, slug=Identifier(alias).lower(), canonical=alias)
        session.add(nn)
        session.commit()

    for alias in aliases:
        db.unalias_nick(alias)

    for alias in aliases:
        found = session.query(Nicknames) \
                       .filter(Nicknames.nick_id == nick_id) \
                       .all()
        assert len(found) == 1
    session.close()


def test_delete_nick_group(db):
    session = db.ssession()
    aliases = ['Embolalia', 'Embo']
    nick_id = 42
    for alias in aliases:
        nn = Nicknames(nick_id=nick_id, slug=Identifier(alias).lower(), canonical=alias)
        session.add(nn)
        session.commit()

    db.set_nick_value(aliases[0], 'foo', 'bar')
    db.set_nick_value(aliases[1], 'spam', 'eggs')

    db.delete_nick_group(aliases[0])

    # Nothing else has created values, so we know the tables are empty
    nicks = session.query(Nicknames).all()
    assert len(nicks) == 0
    data = session.query(NickValues).first()
    assert data is None
    session.close()


def test_merge_nick_groups(db):
    session = db.ssession()
    aliases = ['Embolalia', 'Embo']
    for nick_id, alias in enumerate(aliases):
        nn = Nicknames(nick_id=nick_id, slug=Identifier(alias).lower(), canonical=alias)
        session.add(nn)
        session.commit()

    finals = (('foo', 'bar'), ('bar', 'blue'), ('spam', 'eggs'))

    db.set_nick_value(aliases[0], finals[0][0], finals[0][1])
    db.set_nick_value(aliases[0], finals[1][0], finals[1][1])
    db.set_nick_value(aliases[1], 'foo', 'baz')
    db.set_nick_value(aliases[1], finals[2][0], finals[2][1])

    db.merge_nick_groups(aliases[0], aliases[1])

    nick_ids = session.query(Nicknames.nick_id).all()
    nick_id = nick_ids[0][0]
    alias_id = nick_ids[1][0]
    assert nick_id == alias_id

    for key, value in finals:
        found = session.query(NickValues.value) \
                       .filter(NickValues.nick_id == nick_id) \
                       .filter(NickValues.key == key) \
                       .scalar()
        assert json.loads(unicode(found)) == value
    session.close()


def test_set_channel_value(db):
    session = db.ssession()
    db.set_channel_value('#asdf', 'qwer', 'zxcv')
    result = session.query(ChannelValues.value) \
                    .filter(ChannelValues.channel == '#asdf') \
                    .filter(ChannelValues.key == 'qwer') \
                    .scalar()
    assert result == '"zxcv"'
    session.close()


def test_delete_channel_value(db):
    db.set_channel_value('#asdf', 'wasd', 'uldr')
    assert db.get_channel_value('#asdf', 'wasd') == 'uldr'
    db.delete_channel_value('#asdf', 'wasd')
    assert db.get_channel_value('#asdf', 'wasd') is None


def test_get_channel_value(db):
    session = db.ssession()

    cv = ChannelValues(channel='#asdf', key='qwer', value='\"zxcv\"')
    session.add(cv)
    session.commit()

    result = db.get_channel_value('#asdf', 'qwer')
    assert result == 'zxcv'
    session.close()


def test_get_channel_value_default(db):
    assert db.get_channel_value("TestChan", "DoesntExist") is None
    assert db.get_channel_value("TestChan", "DoesntExist", "MyDefault") == "MyDefault"


def test_get_nick_or_channel_value(db):
    db.set_nick_value('asdf', 'qwer', 'poiu')
    db.set_channel_value('#asdf', 'qwer', '/.,m')
    assert db.get_nick_or_channel_value('asdf', 'qwer') == 'poiu'
    assert db.get_nick_or_channel_value('#asdf', 'qwer') == '/.,m'


def test_get_nick_or_channel_value_default(db):
    assert db.get_nick_or_channel_value("Test", "DoesntExist") is None
    assert db.get_nick_or_channel_value("Test", "DoesntExist", "MyDefault") == "MyDefault"


def test_get_preferred_value(db):
    db.set_nick_value('asdf', 'qwer', 'poiu')
    db.set_channel_value('#asdf', 'qwer', '/.,m')
    db.set_channel_value('#asdf', 'lkjh', '1234')
    names = ['asdf', '#asdf']
    assert db.get_preferred_value(names, 'qwer') == 'poiu'
    assert db.get_preferred_value(names, 'lkjh') == '1234'


def test_set_plugin_value(db):
    session = db.ssession()
    db.set_plugin_value('plugname', 'qwer', 'zxcv')
    result = session.query(PluginValues.value) \
                    .filter(PluginValues.plugin == 'plugname') \
                    .filter(PluginValues.key == 'qwer') \
                    .scalar()
    assert result == '"zxcv"'
    session.close()


def test_get_plugin_value(db):
    session = db.ssession()

    pv = PluginValues(plugin='plugname', key='qwer', value='\"zxcv\"')
    session.add(pv)
    session.commit()

    result = db.get_plugin_value('plugname', 'qwer')
    assert result == 'zxcv'
    session.close()


def test_get_plugin_value_default(db):
    assert db.get_plugin_value("TestPlugin", "DoesntExist") is None
    assert db.get_plugin_value("TestPlugin", "DoesntExist", "MyDefault") == "MyDefault"


def test_delete_plugin_value(db):
    db.set_plugin_value('plugin', 'wasd', 'uldr')
    assert db.get_plugin_value('plugin', 'wasd') == 'uldr'
    db.delete_plugin_value('plugin', 'wasd')
    assert db.get_plugin_value('plugin', 'wasd') is None
